@@ -25,6 +25,7 @@
 #include <stdio.h>
 #include <unistd.h>
 #include <signal.h>
+#include <sys/time.h>
 
 #include <xcb/bigreq.h>
 #include <xcb/randr.h>
@@ -33,9 +34,6 @@
 #include <xcb/xtest.h>
 #include <xcb/shape.h>
 
-#include <X11/Xlib-xcb.h>
-#include <X11/XKBlib.h>
-
 #include <glib-unix.h>
 
 #include "awesome.h"
@@ -60,6 +58,12 @@ awesome_t globalconf;
 /** argv used to run awesome */
 static char *awesome_argv;
 
+/** time of last main loop wakeup */
+static struct timeval last_wakeup;
+
+/** current limit for the main loop's runtime */
+static float main_loop_iteration_limit = 0.1;
+
 /** Call before exiting.
  */
 void
@@ -68,6 +72,23 @@ awesome_atexit(bool restart)
     lua_pushboolean(globalconf.L, restart);
     signal_object_emit(globalconf.L, &global_signals, "exit", 1);
 
+    /* Move clients where we want them to be and keep the stacking order intact */
+    foreach(c, globalconf.stack)
+    {
+        xcb_reparent_window(globalconf.connection, (*c)->window, globalconf.screen->root,
+                (*c)->geometry.x, (*c)->geometry.y);
+    }
+
+    /* Save the client order.  This is useful also for "hard" restarts. */
+    xcb_window_t *wins = p_alloca(xcb_window_t, globalconf.clients.len);
+    int n = 0;
+    foreach(client, globalconf.clients)
+        wins[n++] = (*client)->window;
+
+    xcb_change_property(globalconf.connection, XCB_PROP_MODE_REPLACE,
+                        globalconf.screen->root,
+                        AWESOME_CLIENT_ORDER, XCB_ATOM_WINDOW, 32, n, wins);
+
     a_dbus_cleanup();
 
     systray_cleanup();
@@ -86,9 +107,40 @@ awesome_atexit(bool restart)
     xcb_aux_sync(globalconf.connection);
 
     /* Disconnect *after* closing lua */
+    xcb_cursor_context_free(globalconf.cursor_ctx);
     xcb_disconnect(globalconf.connection);
 }
 
+/** Restore the client order after a restart */
+static void
+restore_client_order(xcb_get_property_cookie_t prop_cookie)
+{
+    int client_idx = 0;
+    xcb_window_t *windows;
+    xcb_get_property_reply_t *reply;
+
+    reply = xcb_get_property_reply(globalconf.connection, prop_cookie, NULL);
+    if (!reply || reply->format != 32 || reply->value_len == 0) {
+        p_delete(&reply);
+        return;
+    }
+
+    windows = xcb_get_property_value(reply);
+    for (uint32_t i = 0; i < reply->value_len; i++)
+        /* Find windows[i] and swap it to where it belongs */
+        foreach(c, globalconf.clients)
+            if ((*c)->window == windows[i])
+            {
+                client_t *tmp = *c;
+                *c = globalconf.clients.tab[client_idx];
+                globalconf.clients.tab[client_idx] = tmp;
+                client_idx++;
+            }
+
+    luaA_class_emit_signal(globalconf.L, &client_class, "list", 0);
+    p_delete(&reply);
+}
+
 /** Scan X to find windows to manage.
  */
 static void
@@ -99,6 +151,7 @@ scan(xcb_query_tree_cookie_t tree_c)
     xcb_window_t *wins = NULL;
     xcb_get_window_attributes_reply_t *attr_r;
     xcb_get_geometry_reply_t *geom_r;
+    xcb_get_property_cookie_t prop_cookie;
     long state;
 
     tree_r = xcb_query_tree_reply(globalconf.connection,
@@ -108,6 +161,11 @@ scan(xcb_query_tree_cookie_t tree_c)
     if(!tree_r)
         return;
 
+    /* This gets the property and deletes it */
+    prop_cookie = xcb_get_property_unchecked(globalconf.connection, true,
+                          globalconf.screen->root, AWESOME_CLIENT_ORDER,
+                          XCB_ATOM_WINDOW, 0, UINT_MAX);
+
     /* Get the tree of the children windows of the current root window */
     if(!(wins = xcb_query_tree_children(tree_r)))
         fatal("cannot get tree children");
@@ -162,6 +220,8 @@ scan(xcb_query_tree_cookie_t tree_c)
     }
 
     p_delete(&tree_r);
+
+    restore_client_order(prop_cookie);
 }
 
 static void
@@ -183,13 +243,13 @@ a_xcb_check(void)
         else
         {
             uint8_t type = XCB_EVENT_RESPONSE_TYPE(event);
-            if((type == XCB_ENTER_NOTIFY || type == XCB_LEAVE_NOTIFY) && mouse)
+            if(mouse && (type == XCB_ENTER_NOTIFY || type == XCB_LEAVE_NOTIFY
+                        || type == XCB_BUTTON_PRESS || type == XCB_BUTTON_RELEASE))
             {
-                /* Make sure enter/motion/leave events are handled in the
-                 * correct order */
+                /* Make sure enter/motion/leave/press/release events are handled
+                 * in the correct order */
                 event_handle(mouse);
                 p_delete(&mouse);
-                mouse = NULL;
             }
             event_handle(event);
             p_delete(&event);
@@ -209,7 +269,8 @@ a_xcb_io_cb(GIOChannel *source, GIOCondition cond, gpointer data)
     /* a_xcb_check() already handled all events */
 
     if(xcb_connection_has_error(globalconf.connection))
-        fatal("X server connection broke");
+        fatal("X server connection broke (error %d)",
+                xcb_connection_has_error(globalconf.connection));
 
     return TRUE;
 }
@@ -218,9 +279,34 @@ static gint
 a_glib_poll(GPollFD *ufds, guint nfsd, gint timeout)
 {
     guint res;
+    struct timeval now, length_time;
+    float length;
+
+    /* Do all deferred work now */
     awesome_refresh();
+
+    /* Check if the Lua stack is the way it should be */
+    if (lua_gettop(globalconf.L) != 0) {
+        warn("Something was left on the Lua stack, this is a bug!");
+        luaA_dumpstack(globalconf.L);
+        lua_settop(globalconf.L, 0);
+    }
+
+    /* Check how long this main loop iteration took */
+    gettimeofday(&now, NULL);
+    timersub(&now, &last_wakeup, &length_time);
+    length = length_time.tv_sec + length_time.tv_usec * 1.0f / 1e6;
+    if (length > main_loop_iteration_limit) {
+        warn("Last main loop iteration took %.6f seconds! Increasing limit for "
+                "this warning to that value.", length);
+        main_loop_iteration_limit = length;
+    }
+
+    /* Actually do the polling, record time of wakeup and check for new xcb events */
     res = g_poll(ufds, nfsd, timeout);
+    gettimeofday(&last_wakeup, NULL);
     a_xcb_check();
+
     return res;
 }
 
@@ -288,7 +374,7 @@ main(int argc, char **argv)
     ssize_t cmdlen = 1;
     xdgHandle xdg;
     bool no_argb = false;
-    xcb_generic_event_t *event;
+    bool run_test = false;
     xcb_query_tree_cookie_t tree_c;
     static struct option long_options[] =
     {
@@ -300,6 +386,10 @@ main(int argc, char **argv)
         { NULL,      0, NULL, 0 }
     };
 
+    /* Make stdout/stderr line buffered. */
+    setvbuf(stdout, NULL, _IOLBF, 0);
+    setvbuf(stderr, NULL, _IOLBF, 0);
+
     /* clear the globalconf structure */
     p_clear(&globalconf, 1);
     globalconf.keygrabber = LUA_REFNIL;
@@ -340,16 +430,8 @@ main(int argc, char **argv)
             exit_help(EXIT_SUCCESS);
             break;
           case 'k':
-            if(!luaA_parserc(&xdg, confpath, false))
-            {
-                fprintf(stderr, "✘ Configuration file syntax error.\n");
-                return EXIT_FAILURE;
-            }
-            else
-            {
-                fprintf(stderr, "✔ Configuration file syntax OK.\n");
-                return EXIT_SUCCESS;
-            }
+            run_test = true;
+            break;
           case 'c':
             if(a_strlen(optarg))
                 confpath = a_strdup(optarg);
@@ -361,32 +443,40 @@ main(int argc, char **argv)
             break;
         }
 
+    if (run_test)
+    {
+        if(!luaA_parserc(&xdg, confpath, false))
+        {
+            fprintf(stderr, "✘ Configuration file syntax error.\n");
+            return EXIT_FAILURE;
+        }
+        else
+        {
+            fprintf(stderr, "✔ Configuration file syntax OK.\n");
+            return EXIT_SUCCESS;
+        }
+    }
+
     /* register function for signals */
     g_unix_signal_add(SIGINT, exit_on_signal, NULL);
     g_unix_signal_add(SIGTERM, exit_on_signal, NULL);
     g_unix_signal_add(SIGHUP, restart_on_signal, NULL);
 
-    struct sigaction sa = { .sa_handler = signal_fatal, .sa_flags = 0 };
+    struct sigaction sa = { .sa_handler = signal_fatal, .sa_flags = SA_RESETHAND };
     sigemptyset(&sa.sa_mask);
+    sigaction(SIGABRT, &sa, 0);
+    sigaction(SIGBUS, &sa, 0);
+    sigaction(SIGFPE, &sa, 0);
+    sigaction(SIGILL, &sa, 0);
     sigaction(SIGSEGV, &sa, 0);
 
-    /* XLib sucks */
-    XkbIgnoreExtension(True);
-
-    /* X stuff */
-    globalconf.display = XOpenDisplay(NULL);
-    if (globalconf.display == NULL) {
-        fatal("cannot open display");
-    }
-    XSetEventQueueOwner(globalconf.display, XCBOwnsEventQueue);
-    globalconf.default_screen = XDefaultScreen(globalconf.display);
-    globalconf.connection = XGetXCBConnection(globalconf.display);
     /* We have no clue where the input focus is right now */
     globalconf.focus.need_update = true;
 
-    /* Double checking that connection is good and operatable with xcb */
+    /* X stuff */
+    globalconf.connection = xcb_connect(NULL, &globalconf.default_screen);
     if(xcb_connection_has_error(globalconf.connection))
-        fatal("cannot open display");
+        fatal("cannot open display (error %d)", xcb_connection_has_error(globalconf.connection));
 
     globalconf.screen = xcb_aux_get_screen(globalconf.connection, globalconf.default_screen);
     globalconf.default_visual = draw_default_visual(globalconf.screen);
@@ -412,8 +502,8 @@ main(int argc, char **argv)
     xcb_prefetch_extension_data(globalconf.connection, &xcb_xinerama_id);
     xcb_prefetch_extension_data(globalconf.connection, &xcb_shape_id);
 
-    /* Setup the main context */
-    g_main_context_set_poll_func(g_main_context_default(), &a_glib_poll);
+    if (xcb_cursor_context_new(globalconf.connection, globalconf.screen, &globalconf.cursor_ctx) < 0)
+        fatal("Failed to initialize xcb-cursor");
 
     /* initialize dbus */
     a_dbus_init();
@@ -427,43 +517,29 @@ main(int argc, char **argv)
     /* Grab server */
     xcb_grab_server(globalconf.connection);
 
-    /* Make sure there are no pending events. Since we didn't really do anything
-     * at all yet, we will just discard all events which we received so far.
-     * The above GrabServer should make sure no new events are generated. */
-    xcb_aux_sync(globalconf.connection);
-    while ((event = xcb_poll_for_event(globalconf.connection)) != NULL)
-    {
-        /* Make sure errors are printed */
-        uint8_t response_type = XCB_EVENT_RESPONSE_TYPE(event);
-        if(response_type == 0)
-            event_handle(event);
-        p_delete(&event);
-    }
-
     {
         const uint32_t select_input_val = XCB_EVENT_MASK_SUBSTRUCTURE_REDIRECT;
+        xcb_void_cookie_t cookie;
 
         /* This causes an error if some other window manager is running */
-        xcb_change_window_attributes(globalconf.connection,
-                                     globalconf.screen->root,
-                                     XCB_CW_EVENT_MASK, &select_input_val);
+        cookie = xcb_change_window_attributes_checked(globalconf.connection,
+                                                      globalconf.screen->root,
+                                                      XCB_CW_EVENT_MASK, &select_input_val);
+        if (xcb_request_check(globalconf.connection, cookie))
+            fatal("another window manager is already running");
     }
 
-    /* Need to xcb_flush to validate error handler */
-    xcb_aux_sync(globalconf.connection);
-
-    /* Process all errors in the queue if any. There can be no events yet, so if
-     * this function returns something, it must be an error. */
-    if (xcb_poll_for_event(globalconf.connection) != NULL)
-        fatal("another window manager is already running");
-
     /* Prefetch the maximum request length */
     xcb_prefetch_maximum_request_length(globalconf.connection);
 
     /* check for xtest extension */
-    const xcb_query_extension_reply_t *xtest_query;
-    xtest_query = xcb_get_extension_data(globalconf.connection, &xcb_test_id);
-    globalconf.have_xtest = xtest_query->present;
+    const xcb_query_extension_reply_t *query;
+    query = xcb_get_extension_data(globalconf.connection, &xcb_test_id);
+    globalconf.have_xtest = query->present;
+
+    /* check for shape extension */
+    query = xcb_get_extension_data(globalconf.connection, &xcb_shape_id);
+    globalconf.have_shape = query->present;
 
     /* Allocate the key symbols */
     globalconf.keysyms = xcb_key_symbols_alloc(globalconf.connection);
@@ -489,23 +565,27 @@ main(int argc, char **argv)
     spawn_init();
 
     /* The default GC is just a newly created associated with a window with
-     * depth globalconf.default_depth */
-    xcb_window_t tmp_win = xcb_generate_id(globalconf.connection);
+     * depth globalconf.default_depth.
+     * The window_no_focus is used for "nothing has the input focus". */
+    globalconf.focus.window_no_focus = xcb_generate_id(globalconf.connection);
     globalconf.gc = xcb_generate_id(globalconf.connection);
     xcb_create_window(globalconf.connection, globalconf.default_depth,
-                      tmp_win, globalconf.screen->root,
+                      globalconf.focus.window_no_focus, globalconf.screen->root,
                       -1, -1, 1, 1, 0,
                       XCB_COPY_FROM_PARENT, globalconf.visual->visual_id,
-                      XCB_CW_BACK_PIXEL | XCB_CW_BORDER_PIXEL | XCB_CW_COLORMAP,
+                      XCB_CW_BACK_PIXEL | XCB_CW_BORDER_PIXEL |
+                      XCB_CW_OVERRIDE_REDIRECT | XCB_CW_COLORMAP,
                       (const uint32_t [])
                       {
                           globalconf.screen->black_pixel,
                           globalconf.screen->black_pixel,
+                          1,
                           globalconf.default_cmap
                       });
-    xcb_create_gc(globalconf.connection, globalconf.gc, tmp_win, XCB_GC_FOREGROUND | XCB_GC_BACKGROUND,
+    xcb_map_window(globalconf.connection, globalconf.focus.window_no_focus);
+    xcb_create_gc(globalconf.connection, globalconf.gc, globalconf.focus.window_no_focus,
+                  XCB_GC_FOREGROUND | XCB_GC_BACKGROUND,
                   (const uint32_t[]) { globalconf.screen->black_pixel, globalconf.screen->white_pixel });
-    xcb_destroy_window(globalconf.connection, tmp_win);
 
     /* Get the window tree associated to this screen */
     tree_c = xcb_query_tree_unchecked(globalconf.connection,
@@ -518,6 +598,7 @@ main(int argc, char **argv)
 
     /* we will receive events, stop grabbing server */
     xcb_ungrab_server(globalconf.connection);
+    xcb_flush(globalconf.connection);
 
     /* Parse and run configuration file */
     if (!luaA_parserc(&xdg, confpath, true))
@@ -532,9 +613,16 @@ main(int argc, char **argv)
 
     xcb_flush(globalconf.connection);
 
-    /* main event loop */
-    globalconf.loop = g_main_loop_new(NULL, FALSE);
-    g_main_loop_run(globalconf.loop);
+    /* Setup the main context */
+    g_main_context_set_poll_func(g_main_context_default(), &a_glib_poll);
+    gettimeofday(&last_wakeup, NULL);
+
+    /* main event loop (if not NULL, awesome.quit() was already called) */
+    if (globalconf.loop == NULL)
+    {
+        globalconf.loop = g_main_loop_new(NULL, FALSE);
+        g_main_loop_run(globalconf.loop);
+    }
     g_main_loop_unref(globalconf.loop);
     globalconf.loop = NULL;
 
