@@ -36,22 +36,6 @@
 #include <alloca.h>
 #endif
 
-typedef enum
-{
-    East = 0,
-    South,
-    North,
-} orientation_t;
-
-/** A list of possible position, not sex related */
-typedef enum
-{
-    Top = 0,
-    Bottom,
-    Right,
-    Left
-} position_t;
-
 /** \brief replace \c NULL strings with empty strings */
 #define NONULL(x)       (x ? x : "")
 
@@ -96,31 +80,19 @@ typedef enum
         }                                            \
     } while (0)
 
-
-#ifdef __GNUC__
-
-#define p_delete(mem_pp)                           \
-    do {                                           \
-        typeof(**(mem_pp)) **__ptr = (mem_pp);     \
-        free(*__ptr);                              \
-        *__ptr = NULL;                             \
-    } while(0)
-
-#define likely(expr)    __builtin_expect(!!(expr), 1)
-#define unlikely(expr)  __builtin_expect((expr), 0)
-
-#else
-
-#define p_delete(mem_p)                            \
-    do {                                           \
-        void *__ptr = (mem_p);                     \
-        free(*__ptr);                              \
-        *(void **)__ptr = NULL;                    \
+#define p_delete(mem_p)                              \
+    do {                                             \
+        void **__ptr = (void **) (mem_p);            \
+        free(*__ptr);                                \
+        *(void **)__ptr = NULL;                      \
     } while (0)
 
+#ifdef __GNUC__
+#define likely(expr)    __builtin_expect(!!(expr), 1)
+#define unlikely(expr)  __builtin_expect((expr), 0)
+#else
 #define likely(expr)    expr
 #define unlikely(expr)  expr
-
 #endif
 
 static inline void * __attribute__ ((malloc)) xmalloc(ssize_t size)
@@ -250,6 +222,9 @@ static inline int a_strcmp(const char *a, const char *b)
     return strcmp(NONULL(a), NONULL(b));
 }
 
+#define  A_STREQ(a, b) (((a) == (b)) || a_strcmp(a, b) == 0)
+#define A_STRNEQ(a, b) (!A_STREQ(a, b))
+
 /** \brief \c NULL resistant strcasecmp.
  * \param[in]  a     the first string.
  * \param[in]  b     the second string.
@@ -261,6 +236,9 @@ static inline int a_strcasecmp(const char *a, const char *b)
     return strcasecmp(NONULL(a), NONULL(b));
 }
 
+#define  A_STREQ_CASE(a, b) (((a) == (b)) || a_strcasecmp(a, b) == 0)
+#define A_STRNEQ_CASE(a, b) (!A_STREQ_CASE(a, b))
+
 /** \brief \c NULL resistant strncmp.
  * \param[in]  a     the first string.
  * \param[in]  b     the second string.
@@ -273,6 +251,9 @@ static inline int a_strncmp(const char *a, const char *b, ssize_t n)
     return strncmp(NONULL(a), NONULL(b), n);
 }
 
+#define  A_STREQ_N(a, b, n) (((a) == (b)) || (n) == ((ssize_t) 0) || a_strncmp(a, b, n) == 0)
+#define A_STRNEQ_N(a, b) (!A_STREQN(a, b))
+
 ssize_t a_strncpy(char *dst, ssize_t n, const char *src, ssize_t l) __attribute__((nonnull(1)));
 ssize_t a_strcpy(char *dst, ssize_t n, const char *src) __attribute__((nonnull(1)));
 
@@ -338,10 +319,6 @@ void _fatal(int, const char *, const char *, ...)
 void _warn(int, const char *, const char *, ...)
     __attribute__ ((format(printf, 3, 4)));
 
-position_t position_fromstr(const char *, ssize_t);
-const char * position_tostr(position_t);
-orientation_t orientation_fromstr(const char *, ssize_t);
-const char * orientation_tostr(orientation_t);
 void a_exec(const char *);
 
 #endif
