@@ -21,20 +21,20 @@
 
 #include <xcb/xcb.h>
 #include <xcb/randr.h>
+#include <xcb/shape.h>
 #include <xcb/xcb_atom.h>
 #include <xcb/xcb_icccm.h>
 #include <xcb/xcb_event.h>
 
 #include "awesome.h"
 #include "event.h"
-#include "tag.h"
-#include "window.h"
 #include "property.h"
+#include "objects/tag.h"
+#include "objects/drawin.h"
+#include "xwindow.h"
 #include "ewmh.h"
-#include "client.h"
-#include "widget.h"
-#include "titlebar.h"
-#include "key.h"
+#include "objects/client.h"
+#include "keyresolv.h"
 #include "keygrabber.h"
 #include "mousegrabber.h"
 #include "luaa.h"
@@ -128,16 +128,43 @@ event_handle_mousegrabber(int x, int y, uint16_t mask)
     return false;
 }
 
+/** Emit a button signal.
+ * The top of the lua stack has to be the object on which to emit the event.
+ * \param ev The event to handle.
+ */
+static void
+event_emit_button(xcb_button_press_event_t *ev)
+{
+    const char *name;
+    switch(XCB_EVENT_RESPONSE_TYPE(ev))
+    {
+    case XCB_BUTTON_PRESS:
+        name = "button::press";
+        break;
+    case XCB_BUTTON_RELEASE:
+        name = "button::release";
+        break;
+    default:
+        fatal("Invalid event type");
+    }
+
+    /* Push the event's info */
+    lua_pushnumber(globalconf.L, ev->event_x);
+    lua_pushnumber(globalconf.L, ev->event_y);
+    lua_pushnumber(globalconf.L, ev->detail);
+    luaA_pushmodifiers(globalconf.L, ev->state);
+    /* And emit the signal */
+    luaA_object_emit_signal(globalconf.L, -5, name, 4);
+}
+
 /** The button press event handler.
  * \param ev The event.
  */
 static void
 event_handle_button(xcb_button_press_event_t *ev)
 {
-    int screen;
-    const int nb_screen = xcb_setup_roots_length(xcb_get_setup(globalconf.connection));
     client_t *c;
-    wibox_t *wibox;
+    drawin_t *drawin;
 
     globalconf.timestamp = ev->time;
 
@@ -150,57 +177,66 @@ event_handle_button(xcb_button_press_event_t *ev)
      * drop them */
     ev->state &= 0x00ff;
 
-    if((wibox = wibox_getbywin(ev->event))
-       || (wibox = wibox_getbywin(ev->child)))
+    if((drawin = drawin_getbywin(ev->event))
+       || (drawin = drawin_getbywin(ev->child)))
     {
-        /* If the wibox is child, then x,y are
+        /* If the drawin is child, then x,y are
          * relative to root window */
-        if(wibox->window == ev->child)
+        if(drawin->window == ev->child)
         {
-            ev->event_x -= wibox->geometry.x;
-            ev->event_y -= wibox->geometry.y;
+            ev->event_x -= drawin->geometry.x + drawin->border_width;
+            ev->event_y -= drawin->geometry.y + drawin->border_width;
         }
 
-        /* Push the wibox */
-        luaA_object_push(globalconf.L, wibox);
-        /* Duplicate the wibox */
-        lua_pushvalue(globalconf.L, -1);
-        /* Handle the button event on it */
-        event_button_callback(ev, &wibox->buttons, -1, 1, NULL);
-
-        /* then try to match a widget binding */
-        widget_t *w = widget_getbycoords(wibox->orientation, &wibox->widgets,
-                                         wibox->geometry.width,
-                                         wibox->geometry.height,
-                                         &ev->event_x, &ev->event_y);
-        if(w)
-        {
-            /* Push widget (the wibox is already on stack) */
-            luaA_object_push_item(globalconf.L, -1, w);
-            /* Move widget before wibox */
-            lua_insert(globalconf.L, -2);
-            event_button_callback(ev, &w->buttons, -2, 2, NULL);
-        }
-        else
-            /* Remove the wibox, we did not use it */
-            lua_pop(globalconf.L, 1);
-
+        /* Push the drawable */
+        luaA_object_push(globalconf.L, drawin);
+        luaA_object_push_item(globalconf.L, -1, drawin->drawable);
+        /* and handle the button raw button event */
+        event_emit_button(ev);
+        lua_pop(globalconf.L, 1);
+        /* check if any button object matches */
+        event_button_callback(ev, &drawin->buttons, -1, 1, NULL);
+        /* Either we are receiving this due to ButtonPress/Release on the root
+         * window or because we grabbed the button on the window. In the later
+         * case we have to call AllowEvents.
+         * Use AsyncPointer instead of ReplayPointer so that the event is
+         * "eaten" instead of being handled again on the root window.
+         */
+        if(ev->child == XCB_NONE)
+            xcb_allow_events(globalconf.connection,
+                             XCB_ALLOW_ASYNC_POINTER,
+                             XCB_CURRENT_TIME);
     }
-    else if((c = client_getbywin(ev->event)))
+    else if((c = client_getbyframewin(ev->event)))
     {
         luaA_object_push(globalconf.L, c);
+        /* And handle the button raw button event */
+        event_emit_button(ev);
+        /* then check if a titlebar was "hit" */
+        int x = ev->event_x, y = ev->event_y;
+        drawable_t *d = client_get_drawable_offset(c, &x, &y);
+        if (d)
+        {
+            /* Copy the event so that we can fake x/y */
+            xcb_button_press_event_t event = *ev;
+            event.event_x = x;
+            event.event_y = y;
+            luaA_object_push_item(globalconf.L, -1, d);
+            event_emit_button(&event);
+            lua_pop(globalconf.L, 1);
+        }
+        /* then check if any button objects match */
         event_button_callback(ev, &c->buttons, -1, 1, NULL);
         xcb_allow_events(globalconf.connection,
                          XCB_ALLOW_REPLAY_POINTER,
                          XCB_CURRENT_TIME);
     }
     else if(ev->child == XCB_NONE)
-        for(screen = 0; screen < nb_screen; screen++)
-            if(xutil_screen_get(globalconf.connection, screen)->root == ev->event)
-            {
-                event_button_callback(ev, &globalconf.buttons, 0, 0, NULL);
-                return;
-            }
+        if(globalconf.screen->root == ev->event)
+        {
+            event_button_callback(ev, &globalconf.buttons, 0, 0, NULL);
+            return;
+        }
 }
 
 static void
@@ -256,42 +292,76 @@ static void
 event_handle_configurerequest(xcb_configure_request_event_t *ev)
 {
     client_t *c;
-    area_t geometry;
 
     if((c = client_getbywin(ev->window)))
     {
-        geometry = titlebar_geometry_remove(c->titlebar, c->border_width, c->geometry);
+        area_t geometry = c->geometry;
+        uint16_t bw = c->border_width;
+        uint16_t tb_left = c->titlebar[CLIENT_TITLEBAR_LEFT].size;
+        uint16_t tb_right = c->titlebar[CLIENT_TITLEBAR_RIGHT].size;
+        uint16_t tb_top = c->titlebar[CLIENT_TITLEBAR_TOP].size;
+        uint16_t tb_bottom = c->titlebar[CLIENT_TITLEBAR_BOTTOM].size;
+        uint16_t deco_left = bw + tb_left;
+        uint16_t deco_right = bw + tb_right;
+        uint16_t deco_top = bw + tb_top;
+        uint16_t deco_bottom = bw + tb_bottom;
+        int16_t diff_w = 0, diff_h = 0, diff_border = 0;
 
         if(ev->value_mask & XCB_CONFIG_WINDOW_X)
+        {
+            int16_t diff = 0;
             geometry.x = ev->x;
+            xwindow_translate_for_gravity(c->size_hints.win_gravity, deco_left, 0, deco_right, 0, &diff, NULL);
+            geometry.x += diff;
+        }
         if(ev->value_mask & XCB_CONFIG_WINDOW_Y)
+        {
+            int16_t diff = 0;
             geometry.y = ev->y;
+            xwindow_translate_for_gravity(c->size_hints.win_gravity, 0, deco_top, 0, deco_bottom, NULL, &diff);
+            geometry.y += diff;
+        }
         if(ev->value_mask & XCB_CONFIG_WINDOW_WIDTH)
+        {
+            uint16_t old_w = geometry.width;
             geometry.width = ev->width;
+            /* The ConfigureRequest specifies the size of the client window, we want the frame */
+            geometry.width += tb_left + tb_right;
+            diff_w = geometry.width - old_w;
+        }
         if(ev->value_mask & XCB_CONFIG_WINDOW_HEIGHT)
+        {
+            uint16_t old_h = geometry.height;
             geometry.height = ev->height;
-
+            /* The ConfigureRequest specifies the size of the client window, we want the frame */
+            geometry.height += tb_top + tb_bottom;
+            diff_h = geometry.height - old_h;
+        }
         if(ev->value_mask & XCB_CONFIG_WINDOW_BORDER_WIDTH)
         {
+            diff_border = ev->border_width - bw;
+            diff_h += diff_border;
+            diff_w += diff_border;
+
             luaA_object_push(globalconf.L, c);
-            client_set_border_width(globalconf.L, -1, ev->border_width);
+            window_set_border_width(globalconf.L, -1, ev->border_width);
             lua_pop(globalconf.L, 1);
         }
 
-        /* Clients are not allowed to directly mess with stacking parameters. */
-        ev->value_mask &= ~(XCB_CONFIG_WINDOW_SIBLING |
-                                            XCB_CONFIG_WINDOW_STACK_MODE);
-
-        /** Configure request are sent with size relative to real (internal)
-         * window size, i.e. without titlebars and borders. */
-        geometry = titlebar_geometry_add(c->titlebar, c->border_width, geometry);
+        /* If the client resizes without moving itself, apply window gravity */
+        if(c->size_hints.flags & XCB_ICCCM_SIZE_HINT_P_WIN_GRAVITY)
+        {
+            int16_t diff_x = 0, diff_y = 0;
+            xwindow_translate_for_gravity(c->size_hints.win_gravity, diff_border, diff_border, diff_w, diff_h, &diff_x, &diff_y);
+            if(!(ev->value_mask & XCB_CONFIG_WINDOW_X))
+                geometry.x += diff_x;
+            if(!(ev->value_mask & XCB_CONFIG_WINDOW_Y))
+                geometry.y += diff_y;
+        }
 
         if(!client_resize(c, geometry, false))
-        {
-            /* Resize wasn't officially needed, but we don't want to break expectations. */
-            geometry = titlebar_geometry_remove(c->titlebar, c->border_width, c->geometry);
-            window_configure(c->window, geometry, c->border_width);
-        }
+            /* ICCCM 4.1.5 / 4.2.3, if nothing was changed, send an event saying so */
+            client_send_configure(c);
     }
     else
         event_handle_configurerequest_configure_window(ev);
@@ -303,16 +373,13 @@ event_handle_configurerequest(xcb_configure_request_event_t *ev)
 static void
 event_handle_configurenotify(xcb_configure_notify_event_t *ev)
 {
-    int screen_nbr;
-    const xcb_screen_t *screen;
+    const xcb_screen_t *screen = globalconf.screen;
 
-    for(screen_nbr = 0; screen_nbr < xcb_setup_roots_length(xcb_get_setup(globalconf.connection)); screen_nbr++)
-        if((screen = xutil_screen_get(globalconf.connection, screen_nbr)) != NULL
-           && ev->window == screen->root
-           && (ev->width != screen->width_in_pixels
-               || ev->height != screen->height_in_pixels))
-            /* it's not that we panic, but restart */
-            awesome_restart();
+    if(ev->window == screen->root
+       && (ev->width != screen->width_in_pixels
+           || ev->height != screen->height_in_pixels))
+        /* it's not that we panic, but restart */
+        awesome_restart();
 }
 
 /** The destroy notify event handler.
@@ -324,73 +391,51 @@ event_handle_destroynotify(xcb_destroy_notify_event_t *ev)
     client_t *c;
 
     if((c = client_getbywin(ev->window)))
-        client_unmanage(c);
+        client_unmanage(c, false);
     else
         for(int i = 0; i < globalconf.embedded.len; i++)
             if(globalconf.embedded.tab[i].win == ev->window)
             {
                 xembed_window_array_take(&globalconf.embedded, i);
-                widget_invalidate_bytype(widget_systray);
+                luaA_systray_invalidate();
             }
 }
 
-/** Handle a motion notify over widgets.
- * \param wibox The wibox.
- * \param mouse_over The pointer to the registered mouse over widget.
- * \param widget The new widget the mouse is over.
+/** Record that the given drawable contains the pointer.
  */
 static void
-event_handle_widget_motionnotify(wibox_t *wibox,
-                                 widget_t *widget)
+event_drawable_under_mouse(lua_State *L, int ud)
 {
-    if(widget != wibox->mouse_over)
+    void *d;
+
+    lua_pushvalue(L, ud);
+    d = luaA_object_ref(L, -1);
+
+    if (d == globalconf.drawable_under_mouse)
     {
-        if(wibox->mouse_over)
-        {
-            /* Emit mouse leave signal on old widget:
-             * - Push wibox.*/
-            luaA_object_push(globalconf.L, wibox);
-            /* - Push the widget the mouse was on */
-            luaA_object_push_item(globalconf.L, -1, wibox->mouse_over);
-            /* - Invert wibox/widget */
-            lua_insert(globalconf.L, -2);
-            /* - Emit the signal mouse::leave with the wibox as argument */
-            luaA_object_emit_signal(globalconf.L, -2, "mouse::leave", 1);
-            /* - Remove the widget */
-            lua_pop(globalconf.L, 1);
+        /* Nothing to do */
+        luaA_object_unref(L, d);
+        return;
+    }
 
-            /* Re-push wibox */
-            luaA_object_push(globalconf.L, wibox);
-            luaA_object_unref_item(globalconf.L, -1, wibox->mouse_over);
-            /* Remove the wibox */
-            lua_pop(globalconf.L, 1);
-        }
-        if(widget)
-        {
-            /* Get a ref on this widget so that it can't be unref'd from
-             * underneath us (-> invalid pointer dereference):
-             * - Push the wibox */
-            luaA_object_push(globalconf.L, wibox);
-            /* - Push the widget */
-            luaA_object_push_item(globalconf.L, -1, widget);
-            /* - Reference the widget into the wibox */
-            luaA_object_ref_item(globalconf.L, -2, -1);
+    if (globalconf.drawable_under_mouse != NULL)
+    {
+        /* Emit leave on previous drawable */
+        luaA_object_push(L, globalconf.drawable_under_mouse);
+        luaA_object_emit_signal(L, -1, "mouse::leave", 0);
+        lua_pop(L, 1);
 
-            /* Emit mouse::enter signal on new widget:
-             * - Push the widget */
-            luaA_object_push_item(globalconf.L, -1, widget);
-            /* - Move the widget before the wibox we pushed just above */
-            lua_insert(globalconf.L, -2);
-            /* - Emit the signal with the wibox as argument */
-
-            luaA_object_emit_signal(globalconf.L, -2, "mouse::enter", 1);
-            /* - Remove the widget from the stack */
-            lua_pop(globalconf.L, 1);
-        }
-
-        /* Store that this widget was the one with the mouse over */
-        wibox->mouse_over = widget;
+        /* Unref the previous drawable */
+        luaA_object_unref(L, globalconf.drawable_under_mouse);
+        globalconf.drawable_under_mouse = NULL;
+    }
+    if (d != NULL)
+    {
+        /* Reference the drawable for leave event later */
+        globalconf.drawable_under_mouse = d;
 
+        /* Emit enter */
+        luaA_object_emit_signal(L, ud, "mouse::enter", 0);
     }
 }
 
@@ -400,21 +445,45 @@ event_handle_widget_motionnotify(wibox_t *wibox,
 static void
 event_handle_motionnotify(xcb_motion_notify_event_t *ev)
 {
-    wibox_t *wibox;
+    drawin_t *w;
+    client_t *c;
 
     globalconf.timestamp = ev->time;
 
     if(event_handle_mousegrabber(ev->root_x, ev->root_y, ev->state))
         return;
 
-    if((wibox = wibox_getbywin(ev->event)))
+    if((c = client_getbyframewin(ev->event)))
     {
-        widget_t *w = widget_getbycoords(wibox->orientation, &wibox->widgets,
-                                         wibox->geometry.width,
-                                         wibox->geometry.height,
-                                         &ev->event_x, &ev->event_y);
+        luaA_object_push(globalconf.L, c);
+        lua_pushnumber(globalconf.L, ev->event_x);
+        lua_pushnumber(globalconf.L, ev->event_y);
+        luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
 
-        event_handle_widget_motionnotify(wibox, w);
+        /* now check if a titlebar was "hit" */
+        int x = ev->event_x, y = ev->event_y;
+        drawable_t *d = client_get_drawable_offset(c, &x, &y);
+        if (d)
+        {
+            luaA_object_push_item(globalconf.L, -1, d);
+            event_drawable_under_mouse(globalconf.L, -1);
+            lua_pushnumber(globalconf.L, x);
+            lua_pushnumber(globalconf.L, y);
+            luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
+            lua_pop(globalconf.L, 1);
+        }
+        lua_pop(globalconf.L, 1);
+    }
+
+    if((w = drawin_getbywin(ev->event)))
+    {
+        luaA_object_push(globalconf.L, w);
+        luaA_object_push_item(globalconf.L, -1, w->drawable);
+        event_drawable_under_mouse(globalconf.L, -1);
+        lua_pushnumber(globalconf.L, ev->event_x);
+        lua_pushnumber(globalconf.L, ev->event_y);
+        luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
+        lua_pop(globalconf.L, 2);
     }
 }
 
@@ -424,7 +493,6 @@ event_handle_motionnotify(xcb_motion_notify_event_t *ev)
 static void
 event_handle_leavenotify(xcb_leave_notify_event_t *ev)
 {
-    wibox_t *wibox;
     client_t *c;
 
     globalconf.timestamp = ev->time;
@@ -432,40 +500,19 @@ event_handle_leavenotify(xcb_leave_notify_event_t *ev)
     if(ev->mode != XCB_NOTIFY_MODE_NORMAL)
         return;
 
-    if((c = client_getbytitlebarwin(ev->event)) || (c = client_getbywin(ev->event)))
+    /* Ignore leave with detail inferior (we were left for a window contained in
+     * our window, so technically the pointer is still inside of this window).
+     */
+    if(ev->detail != XCB_NOTIFY_DETAIL_INFERIOR && (c = client_getbyframewin(ev->event)))
     {
-        if(globalconf.hooks.mouse_leave != LUA_REFNIL)
-        {
-            luaA_object_push(globalconf.L, c);
-            luaA_dofunction_from_registry(globalconf.L, globalconf.hooks.mouse_leave, 1, 0);
-        }
         luaA_object_push(globalconf.L, c);
         luaA_object_emit_signal(globalconf.L, -1, "mouse::leave", 0);
         lua_pop(globalconf.L, 1);
     }
 
-    if((wibox = wibox_getbywin(ev->event)))
-    {
-        if(wibox->mouse_over)
-        {
-            /* Push the wibox */
-            luaA_object_push(globalconf.L, wibox);
-            /* Push the widget the mouse is over in this wibox */
-            luaA_object_push_item(globalconf.L, -1, wibox->mouse_over);
-            /* Move the widget before the wibox */
-            lua_insert(globalconf.L, -2);
-            /* Emit mouse::leave signal on widget the mouse was over with
-             * its wibox as argument */
-            luaA_object_emit_signal(globalconf.L, -2, "mouse::leave", 1);
-            /* Remove the widget from the stack */
-            lua_pop(globalconf.L, 1);
-            wibox_clear_mouse_over(wibox);
-        }
-
-        luaA_object_push(globalconf.L, wibox);
-        luaA_object_emit_signal(globalconf.L, -1, "mouse::leave", 0);
-        lua_pop(globalconf.L, 1);
-    }
+    lua_pushnil(globalconf.L);
+    event_drawable_under_mouse(globalconf.L, -1);
+    lua_pop(globalconf.L, 1);
 }
 
 /** The enter notify event handler.
@@ -475,39 +522,45 @@ static void
 event_handle_enternotify(xcb_enter_notify_event_t *ev)
 {
     client_t *c;
-    wibox_t *wibox;
+    drawin_t *drawin;
 
     globalconf.timestamp = ev->time;
 
     if(ev->mode != XCB_NOTIFY_MODE_NORMAL)
         return;
 
-    if((wibox = wibox_getbywin(ev->event)))
+    if((drawin = drawin_getbywin(ev->event)))
     {
-        widget_t *w = widget_getbycoords(wibox->orientation, &wibox->widgets,
-                                         wibox->geometry.width,
-                                         wibox->geometry.height,
-                                         &ev->event_x, &ev->event_y);
-
-        event_handle_widget_motionnotify(wibox, w);
-
-        luaA_object_push(globalconf.L, wibox);
-        luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
-        lua_pop(globalconf.L, 1);
+        luaA_object_push(globalconf.L, drawin);
+        luaA_object_push_item(globalconf.L, -1, drawin->drawable);
+        event_drawable_under_mouse(globalconf.L, -1);
+        lua_pop(globalconf.L, 2);
     }
 
-    if((c = client_getbytitlebarwin(ev->event))
-       || (c = client_getbywin(ev->event)))
+    if((c = client_getbyframewin(ev->event)))
     {
-        if(globalconf.hooks.mouse_enter != LUA_REFNIL)
-        {
-            luaA_object_push(globalconf.L, c);
-            luaA_dofunction_from_registry(globalconf.L, globalconf.hooks.mouse_enter, 1, 0);
-        }
         luaA_object_push(globalconf.L, c);
-        luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
+        /* Ignore enter with detail inferior: The pointer was previously inside
+         * of a child window, so technically this isn't a 'real' enter.
+         */
+        if (ev->detail != XCB_NOTIFY_DETAIL_INFERIOR)
+            luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
+
+        drawable_t *d = client_get_drawable(c, ev->event_x, ev->event_y);
+        if (d)
+        {
+            luaA_object_push_item(globalconf.L, -1, d);
+            event_drawable_under_mouse(globalconf.L, -1);
+            lua_pop(globalconf.L, 1);
+        }
         lua_pop(globalconf.L, 1);
     }
+    else if (ev->event == globalconf.screen->root) {
+        /* When there are multiple X screens with awesome running separate
+         * instances, reset focus.
+         */
+        globalconf.focus.need_update = true;
+    }
 }
 
 /** The focus in event handler.
@@ -516,7 +569,10 @@ event_handle_enternotify(xcb_enter_notify_event_t *ev)
 static void
 event_handle_focusin(xcb_focus_in_event_t *ev)
 {
-    client_t *c;
+    if (ev->mode == XCB_NOTIFY_MODE_GRAB
+            || ev->mode == XCB_NOTIFY_MODE_UNGRAB)
+        /* Ignore focus changes due to keyboard grabs */
+        return;
 
     /* Events that we are interested in: */
     switch(ev->detail)
@@ -531,9 +587,15 @@ event_handle_focusin(xcb_focus_in_event_t *ev)
          */
         case XCB_NOTIFY_DETAIL_NONLINEAR_VIRTUAL:
         case XCB_NOTIFY_DETAIL_NONLINEAR:
-            if((c = client_getbytitlebarwin(ev->event))
-               || (c = client_getbywin(ev->event)))
+          {
+            client_t *c;
+
+            if((c = client_getbywin(ev->event))) {
+                /* If there is still a pending focus change, do it now. */
+                client_focus_refresh();
                 client_focus_update(c);
+            }
+          }
         /* all other events are ignored */
         default:
             break;
@@ -546,15 +608,15 @@ event_handle_focusin(xcb_focus_in_event_t *ev)
 static void
 event_handle_expose(xcb_expose_event_t *ev)
 {
-    wibox_t *wibox;
+    drawin_t *drawin;
+    client_t *client;
 
-    /* If the wibox got need_update set, skip this because it will be repainted
-     * soon anyway. Without this we could be painting garbage to the screen!
-     */
-    if((wibox = wibox_getbywin(ev->window)) && !wibox->need_update)
-        wibox_refresh_pixmap_partial(wibox,
-                                     ev->x, ev->y,
-                                     ev->width, ev->height);
+    if((drawin = drawin_getbywin(ev->window)))
+        drawin_refresh_pixmap_partial(drawin,
+                                      ev->x, ev->y,
+                                      ev->width, ev->height);
+    if ((client = client_getbyframewin(ev->window)))
+        client_refresh_partial(client, ev->x, ev->y, ev->width, ev->height);
 }
 
 /** The key press event handler.
@@ -575,17 +637,15 @@ event_handle_key(xcb_key_press_event_t *ev)
                 warn("error running function: %s", lua_tostring(globalconf.L, -1));
                 luaA_keygrabber_stop(globalconf.L);
             }
-            else if(!lua_isboolean(globalconf.L, -1) || !lua_toboolean(globalconf.L, -1))
-                luaA_keygrabber_stop(globalconf.L);
         }
         lua_pop(globalconf.L, 1);  /* pop returned value or function if not called */
     }
     else
     {
         /* get keysym ignoring all modifiers */
-        xcb_keysym_t keysym = key_getkeysym(ev->detail, 0);
+        xcb_keysym_t keysym = keyresolv_get_keysym(ev->detail, 0);
         client_t *c;
-        if((c = client_getbywin(ev->event)))
+        if((c = client_getbywin(ev->event)) || (c = client_getbynofocuswin(ev->event)))
         {
             luaA_object_push(globalconf.L, c);
             event_key_callback(ev, &c->keys, -1, 1, &keysym);
@@ -601,7 +661,6 @@ event_handle_key(xcb_key_press_event_t *ev)
 static void
 event_handle_maprequest(xcb_map_request_event_t *ev)
 {
-    int phys_screen;
     client_t *c;
     xcb_get_window_attributes_cookie_t wa_c;
     xcb_get_window_attributes_reply_t *wa_r;
@@ -624,7 +683,7 @@ event_handle_maprequest(xcb_map_request_event_t *ev)
     else if((c = client_getbywin(ev->window)))
     {
         /* Check that it may be visible, but not asked to be hidden */
-        if(client_maybevisible(c, c->screen) && !c->hidden)
+        if(client_maybevisible(c) && !c->hidden)
         {
             luaA_object_push(globalconf.L, c);
             client_set_minimized(globalconf.L, -1, false);
@@ -642,9 +701,7 @@ event_handle_maprequest(xcb_map_request_event_t *ev)
             goto bailout;
         }
 
-        phys_screen = xutil_root2screen(globalconf.connection, geom_r->root);
-
-        client_manage(ev->window, geom_r, phys_screen, false);
+        client_manage(ev->window, geom_r, false);
 
         p_delete(&geom_r);
     }
@@ -662,21 +719,14 @@ event_handle_unmapnotify(xcb_unmap_notify_event_t *ev)
     client_t *c;
 
     if((c = client_getbywin(ev->window)))
-    {
-        if(ev->event == xutil_screen_get(globalconf.connection, c->phys_screen)->root
-           && XCB_EVENT_SENT(ev))
-        {
-            client_unmanage(c);
-            xcb_unmap_window(globalconf.connection, ev->window);
-        }
-    }
+        client_unmanage(c, true);
     else
         for(int i = 0; i < globalconf.embedded.len; i++)
             if(globalconf.embedded.tab[i].win == ev->window)
             {
                 xembed_window_array_take(&globalconf.embedded, i);
-                widget_invalidate_bytype(widget_systray);
                 xcb_change_save_set(globalconf.connection, XCB_SET_MODE_DELETE, ev->window);
+                luaA_systray_invalidate();
             }
 }
 
@@ -706,6 +756,24 @@ event_handle_randr_screen_change_notify(xcb_randr_screen_change_notify_event_t *
     awesome_restart();
 }
 
+/** The shape notify event handler.
+ * \param ev The event.
+ */
+static void
+event_handle_shape_notify(xcb_shape_notify_event_t *ev)
+{
+    client_t *c = client_getbywin(ev->affected_window);
+    if (c)
+    {
+        luaA_object_push(globalconf.L, c);
+        if (ev->shape_kind == XCB_SHAPE_SK_BOUNDING)
+            luaA_object_emit_signal(globalconf.L, -1, "property::shape_client_bounding", 0);
+        if (ev->shape_kind == XCB_SHAPE_SK_CLIP)
+            luaA_object_emit_signal(globalconf.L, -1, "property::shape_client_clip", 0);
+        lua_pop(globalconf.L, 1);
+    }
+}
+
 /** The client message event handler.
  * \param ev The event.
  */
@@ -757,22 +825,19 @@ event_handle_mappingnotify(xcb_mapping_notify_event_t *ev)
                             &globalconf.shiftlockmask, &globalconf.capslockmask,
                             &globalconf.modeswitchmask);
 
-        int nscreen = xcb_setup_roots_length(xcb_get_setup(globalconf.connection));
-
         /* regrab everything */
-        for(int phys_screen = 0; phys_screen < nscreen; phys_screen++)
-        {
-            xcb_screen_t *s = xutil_screen_get(globalconf.connection, phys_screen);
-            /* yes XCB_BUTTON_MASK_ANY is also for grab_key even if it's look weird */
-            xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, s->root, XCB_BUTTON_MASK_ANY);
-            window_grabkeys(s->root, &globalconf.keys);
-        }
+        xcb_screen_t *s = globalconf.screen;
+        /* yes XCB_BUTTON_MASK_ANY is also for grab_key even if it's look weird */
+        xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, s->root, XCB_BUTTON_MASK_ANY);
+        xwindow_grabkeys(s->root, &globalconf.keys);
 
         foreach(_c, globalconf.clients)
         {
             client_t *c = *_c;
             xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, c->window, XCB_BUTTON_MASK_ANY);
-            window_grabkeys(c->window, &c->keys);
+            xwindow_grabkeys(c->window, &c->keys);
+            if (c->nofocus_window)
+                xwindow_grabkeys(c->nofocus_window, &c->keys);
         }
     }
 }
@@ -782,8 +847,13 @@ event_handle_reparentnotify(xcb_reparent_notify_event_t *ev)
 {
     client_t *c;
 
-    if((c = client_getbywin(ev->window)))
-        client_unmanage(c);
+    if((c = client_getbywin(ev->window)) && c->frame_window != ev->parent)
+    {
+        /* Ignore reparents to the root window, they *might* be caused by
+         * ourselves if a client quickly unmaps and maps itself again. */
+        if (ev->parent != globalconf.screen->root)
+            client_unmanage(c, true);
+    }
 }
 
 /** \brief awesome xerror function.
@@ -800,8 +870,8 @@ xerror(xcb_generic_error_t *e)
            && e->major_code == XCB_SET_INPUT_FOCUS)
        || (e->error_code == XCB_VALUE
            && e->major_code == XCB_KILL_CLIENT)
-       || (e->major_code == XCB_CONFIGURE_WINDOW
-           && e->error_code == XCB_MATCH))
+       || (e->error_code == XCB_MATCH
+           && e->major_code == XCB_CONFIGURE_WINDOW))
         return;
 
     warn("X error: request=%s (major %d, minor %d), error=%s (%d)",
@@ -849,6 +919,7 @@ void event_handle(xcb_generic_event_t *event)
     }
 
     static uint8_t randr_screen_change_notify = 0;
+    static uint8_t shape_notify = 0;
 
     if(randr_screen_change_notify == 0)
     {
@@ -858,9 +929,19 @@ void event_handle(xcb_generic_event_t *event)
         if(randr_query->present)
             randr_screen_change_notify = randr_query->first_event + XCB_RANDR_SCREEN_CHANGE_NOTIFY;
     }
+    if(shape_notify == 0)
+    {
+        /* check for shape extension */
+        const xcb_query_extension_reply_t *shape_query;
+        shape_query = xcb_get_extension_data(globalconf.connection, &xcb_shape_id);
+        if(shape_query->present)
+            shape_notify = shape_query->first_event + XCB_SHAPE_NOTIFY;
+    }
 
     if (response_type == randr_screen_change_notify)
         event_handle_randr_screen_change_notify((void *) event);
+    if (response_type == shape_notify)
+        event_handle_shape_notify((void *) event);
 }
 
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
