@@ -27,6 +27,8 @@
 
 #include <basedir_fs.h>
 
+#include <xcb/xcb_atom.h>
+
 #include "awesome.h"
 #include "config.h"
 #include "objects/timer.h"
@@ -40,6 +42,7 @@
 #include "objects/drawable.h"
 #include "screen.h"
 #include "event.h"
+#include "property.h"
 #include "selection.h"
 #include "systray.h"
 #include "common/xcursor.h"
@@ -60,6 +63,43 @@ extern const struct luaL_Reg awesome_screen_meta[];
 /** Path to config file */
 static char *conffile;
 
+/** Check whether a composite manager is running.
+ * \return True if such a manager is running.
+ */
+static bool
+composite_manager_running(void)
+{
+    xcb_intern_atom_reply_t *atom_r;
+    xcb_get_selection_owner_reply_t *selection_r;
+    char *atom_name;
+    bool result;
+
+    if(!(atom_name = xcb_atom_name_by_screen("_NET_WM_CM", globalconf.default_screen)))
+    {
+        warn("error getting composite manager atom");
+        return false;
+    }
+
+    atom_r = xcb_intern_atom_reply(globalconf.connection,
+                                   xcb_intern_atom_unchecked(globalconf.connection, false,
+                                                             a_strlen(atom_name), atom_name),
+                                   NULL);
+    p_delete(&atom_name);
+    if(!atom_r)
+        return false;
+
+    selection_r = xcb_get_selection_owner_reply(globalconf.connection,
+                                                xcb_get_selection_owner_unchecked(globalconf.connection,
+                                                                                  atom_r->atom),
+                                                NULL);
+    p_delete(&atom_r);
+
+    result = selection_r != NULL && selection_r->owner != XCB_NONE;
+    p_delete(&selection_r);
+
+    return result;
+}
+
 /** Quit awesome.
  * \param L The Lua VM state.
  * \return The number of elements pushed on stack.
@@ -67,6 +107,8 @@ static char *conffile;
 static int
 luaA_quit(lua_State *L)
 {
+    if (globalconf.loop == NULL)
+        globalconf.loop = g_main_loop_new(NULL, FALSE);
     g_main_loop_quit(globalconf.loop);
     return 0;
 }
@@ -128,124 +170,6 @@ luaA_mbstrlen(lua_State *L)
     return 1;
 }
 
-/** Overload standard Lua next function to use __next key on metatable.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaAe_next(lua_State *L)
-{
-    if(luaL_getmetafield(L, 1, "__next"))
-    {
-        lua_insert(L, 1);
-        lua_call(L, lua_gettop(L) - 1, LUA_MULTRET);
-        return lua_gettop(L);
-    }
-
-    luaL_checktype(L, 1, LUA_TTABLE);
-    lua_settop(L, 2);
-    if(lua_next(L, 1))
-        return 2;
-    lua_pushnil(L);
-    return 1;
-}
-
-/** Overload lua_next() function by using __next metatable field
- * to get next elements.
- * \param L The Lua VM stack.
- * \param idx The index number of elements in stack.
- * \return 1 if more elements to come, 0 otherwise.
- */
-static int
-luaA_next(lua_State *L, int idx)
-{
-    if(luaL_getmetafield(L, idx, "__next"))
-    {
-        /* if idx is relative, reduce it since we got __next */
-        if(idx < 0) idx--;
-        /* copy table and then move key */
-        lua_pushvalue(L, idx);
-        lua_pushvalue(L, -3);
-        lua_remove(L, -4);
-        lua_pcall(L, 2, 2, 0);
-        /* next returned nil, it's the end */
-        if(lua_isnil(L, -1))
-        {
-            /* remove nil */
-            lua_pop(L, 2);
-            return 0;
-        }
-        return 1;
-    }
-    else if(lua_istable(L, idx))
-        return lua_next(L, idx);
-    /* remove the key */
-    lua_pop(L, 1);
-    return 0;
-}
-
-/** Generic pairs function.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaA_generic_pairs(lua_State *L)
-{
-    lua_pushvalue(L, lua_upvalueindex(1));  /* return generator, */
-    lua_pushvalue(L, 1);  /* state, */
-    lua_pushnil(L);  /* and initial value */
-    return 3;
-}
-
-/** Overload standard pairs function to use __pairs field of metatables.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaAe_pairs(lua_State *L)
-{
-    if(luaL_getmetafield(L, 1, "__pairs"))
-    {
-        lua_insert(L, 1);
-        lua_call(L, lua_gettop(L) - 1, LUA_MULTRET);
-        return lua_gettop(L);
-    }
-
-    luaL_checktype(L, 1, LUA_TTABLE);
-    return luaA_generic_pairs(L);
-}
-
-static int
-luaA_ipairs_aux(lua_State *L)
-{
-    int i = luaL_checkint(L, 2) + 1;
-    luaL_checktype(L, 1, LUA_TTABLE);
-    lua_pushinteger(L, i);
-    lua_rawgeti(L, 1, i);
-    return (lua_isnil(L, -1)) ? 0 : 2;
-}
-
-/** Overload standard ipairs function to use __ipairs field of metatables.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaAe_ipairs(lua_State *L)
-{
-    if(luaL_getmetafield(L, 1, "__ipairs"))
-    {
-        lua_insert(L, 1);
-        lua_call(L, lua_gettop(L) - 1, LUA_MULTRET);
-        return lua_gettop(L);
-    }
-
-    luaL_checktype(L, 1, LUA_TTABLE);
-    lua_pushvalue(L, lua_upvalueindex(1));
-    lua_pushvalue(L, 1);
-    lua_pushinteger(L, 0);  /* and initial value */
-    return 3;
-}
-
 /** Enhanced type() function which recognize awesome objects.
  * \param L The Lua VM state.
  * \return The number of arguments pushed on the stack.
@@ -269,17 +193,6 @@ luaA_fixups(lua_State *L)
     lua_pushcfunction(L, luaA_mbstrlen);
     lua_setfield(L, -2, "wlen");
     lua_pop(L, 1);
-    /* replace next */
-    lua_pushcfunction(L, luaAe_next);
-    lua_setglobal(L, "next");
-    /* replace pairs */
-    lua_pushcfunction(L, luaAe_next);
-    lua_pushcclosure(L, luaAe_pairs, 1); /* pairs get next as upvalue */
-    lua_setglobal(L, "pairs");
-    /* replace ipairs */
-    lua_pushcfunction(L, luaA_ipairs_aux);
-    lua_pushcclosure(L, luaAe_ipairs, 1);
-    lua_setglobal(L, "ipairs");
     /* replace type */
     lua_pushcfunction(L, luaAe_type);
     lua_setglobal(L, "type");
@@ -288,108 +201,16 @@ luaA_fixups(lua_State *L)
     lua_setglobal(L, "selection");
 }
 
-/** Look for an item: table, function, etc.
- * \param L The Lua VM state.
- * \param item The pointer item.
- */
-bool
-luaA_hasitem(lua_State *L, const void *item)
-{
-    lua_pushnil(L);
-    while(luaA_next(L, -2))
-    {
-        if(lua_topointer(L, -1) == item)
-        {
-            /* remove value and key */
-            lua_pop(L, 2);
-            return true;
-        }
-        if(lua_istable(L, -1))
-            if(luaA_hasitem(L, item))
-            {
-                /* remove key and value */
-                lua_pop(L, 2);
-                return true;
-            }
-        /* remove value */
-        lua_pop(L, 1);
-    }
-    return false;
-}
-
-/** Browse a table pushed on top of the index, and put all its table and
- * sub-table into an array.
- * \param L The Lua VM state.
- * \param elems The elements array.
- * \return False if we encounter an elements already in list.
- */
-static bool
-luaA_isloop_check(lua_State *L, cptr_array_t *elems)
-{
-    if(lua_istable(L, -1))
-    {
-        const void *object = lua_topointer(L, -1);
-
-        /* Check that the object table is not already in the list */
-        for(int i = 0; i < elems->len; i++)
-            if(elems->tab[i] == object)
-                return false;
-
-        /* push the table in the elements list */
-        cptr_array_append(elems, object);
-
-        /* look every object in the "table" */
-        lua_pushnil(L);
-        while(luaA_next(L, -2))
-        {
-            if(!luaA_isloop_check(L, elems))
-            {
-                /* remove key and value */
-                lua_pop(L, 2);
-                return false;
-            }
-            /* remove value, keep key for next iteration */
-            lua_pop(L, 1);
-        }
-    }
-    return true;
-}
-
-/** Check if a table is a loop. When using tables as direct acyclic digram,
- * this is useful.
- * \param L The Lua VM state.
- * \param idx The index of the table in the stack
- * \return True if the table loops.
- */
-bool
-luaA_isloop(lua_State *L, int idx)
-{
-    /* elems is an elements array that we will fill with all array we
-     * encounter while browsing the tables */
-    cptr_array_t elems;
-
-    cptr_array_init(&elems);
-
-    /* push table on top */
-    lua_pushvalue(L, idx);
-
-    bool ret = luaA_isloop_check(L, &elems);
-
-    /* remove pushed table */
-    lua_pop(L, 1);
-
-    cptr_array_wipe(&elems);
-
-    return !ret;
-}
-
 /** awesome global table.
  * \param L The Lua VM state.
  * \return The number of elements pushed on stack.
  * \luastack
- * \lfield font The default font.
- * \lfield font_height The default font height.
  * \lfield conffile The configuration file which has been loaded.
+ * \lfield version The version of awesome.
+ * \lfield release The release name of awesome.
+ * \lfield startup True if we are still in startup, false otherwise.
+ * \lfield startup_errors Error message for errors that occured during startup.
+ * \lfield composite_manager_running True if a composite manager is running.
  */
 static int
 luaA_awesome_index(lua_State *L)
@@ -417,13 +238,27 @@ luaA_awesome_index(lua_State *L)
         return 1;
     }
 
-    if(A_STREQ(buf, "startup_errors") && globalconf.startup_errors.len != 0)
+    if(A_STREQ(buf, "startup"))
     {
+        lua_pushboolean(L, globalconf.loop == NULL);
+        return 1;
+    }
+
+    if(A_STREQ(buf, "startup_errors"))
+    {
+        if (globalconf.startup_errors.len == 0)
+            return 0;
         lua_pushstring(L, globalconf.startup_errors.s);
         return 1;
     }
 
-    return 0;
+    if(A_STREQ(buf, "composite_manager_running"))
+    {
+        lua_pushboolean(L, composite_manager_running());
+        return 1;
+    }
+
+    return luaA_default_index(L);
 }
 
 /** Add a global signal.
@@ -526,7 +361,11 @@ luaA_init(xdgHandle* xdg)
         { "emit_signal", luaA_awesome_emit_signal },
         { "systray", luaA_systray },
         { "load_image", luaA_load_image },
+        { "register_xproperty", luaA_register_xproperty },
+        { "set_xproperty", luaA_set_xproperty },
+        { "get_xproperty", luaA_get_xproperty },
         { "__index", luaA_awesome_index },
+        { "__newindex", luaA_default_newindex },
         { NULL, NULL }
     };
 
@@ -659,37 +498,41 @@ luaA_startup_error(const char *err)
 static bool
 luaA_loadrc(const char *confpath, bool run)
 {
-    if(!luaL_loadfile(globalconf.L, confpath))
-    {
-        if(run)
-        {
-            /* Set the conffile right now so it can be used inside the
-             * configuration file. */
-            conffile = a_strdup(confpath);
-            if(lua_pcall(globalconf.L, 0, LUA_MULTRET, 0))
-            {
-                const char *err = lua_tostring(globalconf.L, -1);
-                luaA_startup_error(err);
-                fprintf(stderr, "%s\n", err);
-                /* An error happened, so reset this. */
-                conffile = NULL;
-            }
-            else
-                return true;
-        }
-        else
-        {
-            lua_pop(globalconf.L, 1);
-            return true;
-        }
-    }
-    else
+    if(luaL_loadfile(globalconf.L, confpath))
     {
         const char *err = lua_tostring(globalconf.L, -1);
         luaA_startup_error(err);
         fprintf(stderr, "%s\n", err);
+        return false;
     }
 
+    if(!run)
+    {
+        lua_pop(globalconf.L, 1);
+        return true;
+    }
+
+    /* Set the conffile right now so it can be used inside the
+     * configuration file. */
+    conffile = a_strdup(confpath);
+    /* Move error handling function before function */
+    lua_pushcfunction(globalconf.L, luaA_dofunction_on_error);
+    lua_insert(globalconf.L, -2);
+    if(!lua_pcall(globalconf.L, 0, 0, -2))
+    {
+        /* Pop luaA_dofunction_on_error */
+        lua_pop(globalconf.L, 1);
+        return true;
+    }
+
+    const char *err = lua_tostring(globalconf.L, -1);
+    luaA_startup_error(err);
+    fprintf(stderr, "%s\n", err);
+    /* An error happened, so reset this. */
+    conffile = NULL;
+    /* Pop luaA_dofunction_on_error() and the error message */
+    lua_pop(globalconf.L, 2);
+
     return false;
 }
 
@@ -760,4 +603,16 @@ luaA_emit_refresh()
     signal_object_emit(globalconf.L, &global_signals, "refresh", 0);
 }
 
+int
+luaA_default_index(lua_State *L)
+{
+    return luaA_class_index_miss_property(L, NULL);
+}
+
+int
+luaA_default_newindex(lua_State *L)
+{
+    return luaA_class_newindex_miss_property(L, NULL);
+}
+
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
