@@ -20,7 +20,6 @@
  */
 
 #include <xcb/xcb_atom.h>
-#include <xcb/xcb_image.h>
 #include <xcb/shape.h>
 #include <cairo-xcb.h>
 
@@ -37,6 +36,7 @@
 
 static area_t titlebar_get_area(client_t *c, client_titlebar_t bar);
 static drawable_t *titlebar_get_drawable(lua_State *L, client_t *c, int cl_idx, client_titlebar_t bar);
+static void client_resize_do(client_t *c, area_t geometry, bool force_notice, bool honor_hints);
 
 /** Collect a client.
  * \param L The Lua VM state.
@@ -54,6 +54,7 @@ client_wipe(client_t *c)
     p_delete(&c->alt_icon_name);
     p_delete(&c->name);
     p_delete(&c->alt_name);
+    p_delete(&c->startup_id);
     if(c->icon)
         cairo_surface_destroy(c->icon);
 }
@@ -70,22 +71,8 @@ client_set_urgent(lua_State *L, int cidx, bool urgent)
 
     if(c->urgent != urgent)
     {
-        xcb_get_property_cookie_t hints =
-            xcb_icccm_get_wm_hints_unchecked(globalconf.connection, c->window);
-
         c->urgent = urgent;
 
-        /* update ICCCM hints */
-        xcb_icccm_wm_hints_t wmh;
-        xcb_icccm_get_wm_hints_reply(globalconf.connection, hints, &wmh, NULL);
-
-        if(urgent)
-            wmh.flags |= XCB_ICCCM_WM_HINT_X_URGENCY;
-        else
-            wmh.flags &= ~XCB_ICCCM_WM_HINT_X_URGENCY;
-
-        xcb_icccm_set_wm_hints(globalconf.connection, c->window, &wmh);
-
         luaA_object_emit_signal(L, cidx, "property::urgent", 0);
     }
 }
@@ -113,6 +100,11 @@ DO_CLIENT_SET_PROPERTY(skip_taskbar)
     client_set_##prop(lua_State *L, int cidx, char *value) \
     { \
         client_t *c = luaA_checkudata(L, cidx, &client_class); \
+        if (A_STREQ(c->prop, value)) \
+        { \
+            p_delete(&value); \
+            return; \
+        } \
         p_delete(&c->prop); \
         c->prop = value; \
         luaA_object_emit_signal(L, cidx, "property::" #signal, 0); \
@@ -172,6 +164,16 @@ client_getbywin(xcb_window_t w)
     return NULL;
 }
 
+client_t *
+client_getbynofocuswin(xcb_window_t w)
+{
+    foreach(c, globalconf.clients)
+        if((*c)->nofocus_window == w)
+            return *c;
+
+    return NULL;
+}
+
 /** Get a client by its frame window.
  * \param w The client window to find.
  * \return A client pointer if found, NULL otherwise.
@@ -334,11 +336,25 @@ client_focus(client_t *c)
     globalconf.focus.need_update = true;
 }
 
+static xcb_window_t
+client_get_nofocus_window(client_t *c)
+{
+    if (c->nofocus_window == XCB_NONE) {
+        c->nofocus_window = xcb_generate_id(globalconf.connection);
+        xcb_create_window(globalconf.connection, globalconf.default_depth, c->nofocus_window, c->frame_window,
+                          -2, -2, 1, 1, 0, XCB_COPY_FROM_PARENT, globalconf.visual->visual_id,
+                          0, NULL);
+        xcb_map_window(globalconf.connection, c->nofocus_window);
+        xwindow_grabkeys(c->nofocus_window, &c->keys);
+    }
+    return c->nofocus_window;
+}
+
 void
 client_focus_refresh(void)
 {
     client_t *c = globalconf.focus.client;
-    xcb_window_t win = globalconf.screen->root;
+    xcb_window_t win = globalconf.focus.window_no_focus;
 
     if(!globalconf.focus.need_update)
         return;
@@ -352,10 +368,7 @@ client_focus_refresh(void)
         if(!c->nofocus)
             win = c->window;
         else
-            /* Focus the root window to make sure the previously focused client
-             * doesn't get any input in case WM_TAKE_FOCUS gets ignored.
-             */
-            win = globalconf.screen->root;
+            win = client_get_nofocus_window(c);
 
         if(client_hasproto(c, WM_TAKE_FOCUS))
             xwindow_takefocus(c->window);
@@ -370,6 +383,24 @@ client_focus_refresh(void)
                         win, globalconf.timestamp);
 }
 
+static void
+border_width_callback(client_t *c, uint16_t old_width, uint16_t new_width)
+{
+    if(c->size_hints.flags & XCB_ICCCM_SIZE_HINT_P_WIN_GRAVITY)
+    {
+        area_t geometry = c->geometry;
+        int16_t diff = new_width - old_width;
+        int16_t diff_x = 0, diff_y = 0;
+        xwindow_translate_for_gravity(c->size_hints.win_gravity,
+                                      diff, diff, diff, diff,
+                                      &diff_x, &diff_y);
+        geometry.x += diff_x;
+        geometry.y += diff_y;
+        /* force_notice = true -> inform client about changes */
+        client_resize_do(c, geometry, true, false);
+    }
+}
+
 static void
 client_update_properties(client_t *c)
 {
@@ -436,9 +467,12 @@ client_manage(xcb_window_t w, xcb_get_geometry_reply_t *wgeom, bool startup)
 
     /* Make sure the window is automatically mapped if awesome exits or dies. */
     xcb_change_save_set(globalconf.connection, XCB_SET_MODE_INSERT, w);
+    if (globalconf.have_shape)
+        xcb_shape_select_input(globalconf.connection, w, 1);
 
     client_t *c = client_new(globalconf.L);
     xcb_screen_t *s = globalconf.screen;
+    c->border_width_callback = (void (*) (void *, uint16_t, uint16_t)) border_width_callback;
 
     /* consider the window banned */
     c->isbanned = true;
@@ -448,12 +482,10 @@ client_manage(xcb_window_t w, xcb_get_geometry_reply_t *wgeom, bool startup)
     xcb_create_window(globalconf.connection, globalconf.default_depth, c->frame_window, s->root,
                       wgeom->x, wgeom->y, wgeom->width, wgeom->height,
                       wgeom->border_width, XCB_COPY_FROM_PARENT, globalconf.visual->visual_id,
-                      XCB_CW_BACK_PIXEL | XCB_CW_BORDER_PIXEL | XCB_CW_BIT_GRAVITY
-                      | XCB_CW_WIN_GRAVITY | XCB_CW_OVERRIDE_REDIRECT | XCB_CW_EVENT_MASK
-                      | XCB_CW_COLORMAP,
+                      XCB_CW_BORDER_PIXEL | XCB_CW_BIT_GRAVITY | XCB_CW_WIN_GRAVITY
+                      | XCB_CW_OVERRIDE_REDIRECT | XCB_CW_EVENT_MASK | XCB_CW_COLORMAP,
                       (const uint32_t [])
                       {
-                          globalconf.screen->black_pixel,
                           globalconf.screen->black_pixel,
                           XCB_GRAVITY_NORTH_WEST,
                           XCB_GRAVITY_NORTH_WEST,
@@ -543,24 +575,9 @@ HANDLE_GEOM(height)
     ewmh_client_check_hints(c);
 
     /* Push client in stack */
-    client_raise(c);
+    stack_client_push(c);
 
-    /* Always stay in NORMAL_STATE. Even though iconified seems more
-     * appropriate sometimes. The only possible loss is that clients not using
-     * visibility events may continue to process data (when banned).
-     * Without any exposes or other events the cost should be fairly limited though.
-     *
-     * Some clients may expect the window to be unmapped when STATE_ICONIFIED.
-     * Two conflicting parts of the ICCCM v2.0 (section 4.1.4):
-     *
-     * "Normal -> Iconic - The client should send a ClientMessage event as described later in this section."
-     * (note no explicit mention of unmapping, while Normal->Widthdrawn does mention that)
-     *
-     * "Once a client's window has left the Withdrawn state, the window will be mapped
-     * if it is in the Normal state and the window will be unmapped if it is in the Iconic state."
-     *
-     * At this stage it's just safer to keep it in normal state and avoid confusion.
-     */
+    /* Put the window in normal state. */
     xwindow_set_state(c->window, XCB_ICCCM_WM_STATE_NORMAL);
 
     if(!startup)
@@ -570,9 +587,9 @@ HANDLE_GEOM(height)
             xcb_get_property_reply(globalconf.connection, startup_id_q, NULL);
         /* Say spawn that a client has been started, with startup id as argument */
         char *startup_id = xutil_get_text_property_from_reply(reply);
+        c->startup_id = startup_id;
         p_delete(&reply);
         spawn_start_notify(c, startup_id);
-        p_delete(&startup_id);
     }
 
     luaA_class_emit_signal(globalconf.L, &client_class, "list", 0);
@@ -633,7 +650,7 @@ client_apply_size_hints(client_t *c, area_t geometry)
     /* Size hints are applied to the window without any decoration */
     client_remove_titlebar_geometry(c, &geometry);
 
-    if(c->size_hints.flags & XCB_ICCCM_SIZE_HINT_P_SIZE)
+    if(c->size_hints.flags & XCB_ICCCM_SIZE_HINT_BASE_SIZE)
     {
         basew = c->size_hints.base_width;
         baseh = c->size_hints.base_height;
@@ -652,7 +669,7 @@ client_apply_size_hints(client_t *c, area_t geometry)
         minw = c->size_hints.min_width;
         minh = c->size_hints.min_height;
     }
-    else if(c->size_hints.flags & XCB_ICCCM_SIZE_HINT_P_SIZE)
+    else if(c->size_hints.flags & XCB_ICCCM_SIZE_HINT_BASE_SIZE)
     {
         /* min size is substituted with base size if not specified */
         minw = c->size_hints.base_width;
@@ -730,12 +747,17 @@ client_resize_do(client_t *c, area_t geometry, bool force_notice, bool honor_hin
     bool send_notice = force_notice;
     bool hide_titlebars = c->fullscreen;
     screen_t *new_screen = screen_getbycoord(geometry.x, geometry.y);
+    bool java_is_broken = true;
 
     if (honor_hints)
         geometry = client_apply_size_hints(c, geometry);
 
     if(c->geometry.width == geometry.width
        && c->geometry.height == geometry.height)
+        /* We are moving without changing the size, see ICCCM 4.2.3 */
+        send_notice = true;
+    if(java_is_broken)
+        /* Java strong. Java Hulk. Java make own rules! */
         send_notice = true;
 
     /* Also store geometry including border */
@@ -768,13 +790,10 @@ client_resize_do(client_t *c, area_t geometry, bool force_notice, bool honor_hin
             (uint32_t[]) { real_geometry.x, real_geometry.y, real_geometry.width, real_geometry.height });
 
     if(send_notice)
-        /* We are moving without changing the size, see ICCCM 4.2.3 */
         client_send_configure(c);
 
     client_restore_enterleave_events();
 
-    screen_client_moveto(c, new_screen, false);
-
     luaA_object_push(globalconf.L, c);
     luaA_object_emit_signal(globalconf.L, -1, "property::geometry", 0);
     if (old_geometry.x != geometry.x)
@@ -787,6 +806,8 @@ client_resize_do(client_t *c, area_t geometry, bool force_notice, bool honor_hin
         luaA_object_emit_signal(globalconf.L, -1, "property::height", 0);
     lua_pop(globalconf.L, 1);
 
+    screen_client_moveto(c, new_screen, false);
+
     /* Update all titlebars */
     for (client_titlebar_t bar = CLIENT_TITLEBAR_TOP; bar < CLIENT_TITLEBAR_COUNT; bar++) {
         if (c->titlebar[bar].drawable == NULL && c->titlebar[bar].size == 0)
@@ -803,29 +824,7 @@ client_resize_do(client_t *c, area_t geometry, bool force_notice, bool honor_hin
         area.y += geometry.y;
         if (hide_titlebars)
             area.width = area.height = 0;
-
-        if (old_geometry.width != geometry.width || old_geometry.height != geometry.height ||
-                drawable->geometry.width == 0 || drawable->geometry.height == 0) {
-            /* Get rid of the old state */
-            drawable_unset_surface(drawable);
-            if (c->titlebar[bar].pixmap != XCB_NONE)
-                xcb_free_pixmap(globalconf.connection, c->titlebar[bar].pixmap);
-            c->titlebar[bar].pixmap = XCB_NONE;
-
-            /* And get us some new state */
-            if (c->titlebar[bar].size != 0 && !hide_titlebars)
-            {
-                c->titlebar[bar].pixmap = xcb_generate_id(globalconf.connection);
-                xcb_create_pixmap(globalconf.connection, globalconf.default_depth, c->titlebar[bar].pixmap,
-                                  globalconf.screen->root, area.width, area.height);
-                cairo_surface_t *surface = cairo_xcb_surface_create(globalconf.connection,
-                                                                    c->titlebar[bar].pixmap, globalconf.visual,
-                                                                    area.width, area.height);
-                drawable_set_surface(drawable, -1, surface, area);
-            } else
-                drawable_set_geometry(drawable, -1, area);
-        } else
-            drawable_set_geometry(drawable, -1, area);
+        drawable_set_geometry(drawable, -1, area);
 
         /* Pop the client and the drawable */
         lua_pop(globalconf.L, 2);
@@ -892,9 +891,49 @@ client_set_minimized(lua_State *L, int cidx, bool s)
         c->minimized = s;
         banning_need_update();
         if(s)
+        {
+            /* ICCCM: To transition from ICONIC to NORMAL state, the client
+             * should just map the window. Thus, iconic clients need to be
+             * unmapped, else the MapWindow request doesn't have any effect.
+             */
             xwindow_set_state(c->window, XCB_ICCCM_WM_STATE_ICONIC);
+
+            uint32_t no_event[] = { 0 };
+            const uint32_t client_select_input_val[] = { CLIENT_SELECT_INPUT_EVENT_MASK };
+            const uint32_t frame_select_input_val[] = { FRAME_SELECT_INPUT_EVENT_MASK };
+            xcb_grab_server(globalconf.connection);
+            xcb_change_window_attributes(globalconf.connection,
+                                         globalconf.screen->root,
+                                         XCB_CW_EVENT_MASK,
+                                         no_event);
+            xcb_change_window_attributes(globalconf.connection,
+                                         c->frame_window,
+                                         XCB_CW_EVENT_MASK,
+                                         no_event);
+            xcb_change_window_attributes(globalconf.connection,
+                                         c->window,
+                                         XCB_CW_EVENT_MASK,
+                                         no_event);
+            xcb_unmap_window(globalconf.connection, c->window);
+            xcb_change_window_attributes(globalconf.connection,
+                                         globalconf.screen->root,
+                                         XCB_CW_EVENT_MASK,
+                                         ROOT_WINDOW_EVENT_MASK);
+            xcb_change_window_attributes(globalconf.connection,
+                                         c->frame_window,
+                                         XCB_CW_EVENT_MASK,
+                                         frame_select_input_val);
+            xcb_change_window_attributes(globalconf.connection,
+                                         c->window,
+                                         XCB_CW_EVENT_MASK,
+                                         client_select_input_val);
+            xcb_ungrab_server(globalconf.connection);
+        }
         else
+        {
             xwindow_set_state(c->window, XCB_ICCCM_WM_STATE_NORMAL);
+            xcb_map_window(globalconf.connection, c->window);
+        }
         if(strut_has_value(&c->strut))
             screen_emit_signal(globalconf.L, c->screen, "property::workarea", 0);
         luaA_object_emit_signal(L, cidx, "property::minimized", 0);
@@ -955,8 +994,7 @@ client_set_fullscreen(lua_State *L, int cidx, bool s)
         if(s)
         {
             /* remove any max state */
-            client_set_maximized_horizontal(L, cidx, false);
-            client_set_maximized_vertical(L, cidx, false);
+            client_set_maximized(L, cidx, false);
             /* You can only be part of one of the special layers. */
             client_set_below(L, cidx, false);
             client_set_above(L, cidx, false);
@@ -973,6 +1011,16 @@ client_set_fullscreen(lua_State *L, int cidx, bool s)
     }
 }
 
+/** Get a clients maximized state (horizontally and vertically).
+ * \param c The client.
+ * \return The maximized state.
+ */
+static int
+client_get_maximized(client_t *c)
+{
+    return c->maximized_horizontal && c->maximized_vertical;
+}
+
 /** Set a client horizontally|vertically maximized.
  * \param L The Lua VM state.
  * \param cidx The client index.
@@ -989,9 +1037,12 @@ client_set_fullscreen(lua_State *L, int cidx, bool s)
             if(s) \
                 client_set_fullscreen(L, abs_cidx, false); \
             lua_pushboolean(L, s); \
+            int max_before = client_get_maximized(c); \
             c->maximized_##type = s; \
             luaA_object_emit_signal(L, abs_cidx, "request::maximized_" #type, 1); \
             luaA_object_emit_signal(L, abs_cidx, "property::maximized_" #type, 0); \
+            if(max_before != client_get_maximized(c)) \
+                luaA_object_emit_signal(L, abs_cidx, "property::maximized", 0); \
             stack_windows(); \
         } \
     }
@@ -999,6 +1050,18 @@ DO_FUNCTION_CLIENT_MAXIMIZED(vertical)
 DO_FUNCTION_CLIENT_MAXIMIZED(horizontal)
 #undef DO_FUNCTION_CLIENT_MAXIMIZED
 
+/** Set a client maximized (horizontally and vertically).
+ * \param L The Lua VM state.
+ * \param cidx The client index.
+ * \param s Set or not the client maximized attribute.
+ */
+void
+client_set_maximized(lua_State *L, int cidx, bool s)
+{
+    client_set_maximized_horizontal(L, cidx, s);
+    client_set_maximized_vertical(L, cidx, s);
+}
+
 /** Set a client above, or not.
  * \param L The Lua VM state.
  * \param cidx The client index.
@@ -1119,7 +1182,7 @@ client_unban(client_t *c)
 void
 client_unmanage(client_t *c, bool window_valid)
 {
-    /* Reset transient_for attributes of widows that maybe referring to us */
+    /* Reset transient_for attributes of windows that might be referring to us */
     foreach(_tc, globalconf.clients)
     {
         client_t *tc = *_tc;
@@ -1155,18 +1218,11 @@ client_unmanage(client_t *c, bool window_valid)
         if (c->titlebar[bar].drawable == NULL)
             continue;
 
+        /* Forget about the drawable */
         luaA_object_push(globalconf.L, c);
-        luaA_object_push_item(globalconf.L, -1, c->titlebar[bar].drawable);
-
-        /* Make the drawable unusable */
-        drawable_unset_surface(c->titlebar[bar].drawable);
-        if (c->titlebar[bar].pixmap != XCB_NONE)
-            xcb_free_pixmap(globalconf.connection, c->titlebar[bar].pixmap);
-
-        /* And forget about it */
-        luaA_object_unref_item(globalconf.L, -2, c->titlebar[bar].drawable);
+        luaA_object_unref_item(globalconf.L, -1, c->titlebar[bar].drawable);
         c->titlebar[bar].drawable = NULL;
-        lua_pop(globalconf.L, 2);
+        lua_pop(globalconf.L, 1);
     }
 
     /* Clear our event mask so that we don't receive any events from now on,
@@ -1187,13 +1243,21 @@ client_unmanage(client_t *c, bool window_valid)
         xcb_reparent_window(globalconf.connection, c->window, globalconf.screen->root,
                 c->geometry.x, c->geometry.y);
     }
+
+    /* Ignore all spurious enter/leave notify events */
+    client_ignore_enterleave_events();
+    if (c->nofocus_window != XCB_NONE)
+        xcb_destroy_window(globalconf.connection, c->nofocus_window);
     xcb_destroy_window(globalconf.connection, c->frame_window);
+    client_restore_enterleave_events();
 
     if(window_valid)
     {
         /* Remove this window from the save set since this shouldn't be made visible
          * after a restart anymore. */
         xcb_change_save_set(globalconf.connection, XCB_SET_MODE_DELETE, c->window);
+        if (globalconf.have_shape)
+            xcb_shape_select_input(globalconf.connection, c->window, 0);
 
         /* Do this last to avoid races with clients. According to ICCCM, clients
          * arent allowed to re-use the window until after this. */
@@ -1304,6 +1368,71 @@ client_set_icon(client_t *c, cairo_surface_t *s)
     lua_pop(globalconf.L, 1);
 }
 
+/** Set a client icon.
+ * \param c The client to change.
+ * \param icon A bitmap containing the icon.
+ * \param mask A mask for the bitmap (optional)
+ */
+void
+client_set_icon_from_pixmaps(client_t *c, xcb_pixmap_t icon, xcb_pixmap_t mask)
+{
+    xcb_get_geometry_cookie_t geom_icon_c, geom_mask_c;
+    xcb_get_geometry_reply_t *geom_icon_r, *geom_mask_r = NULL;
+    cairo_surface_t *s_icon, *result;
+
+    geom_icon_c = xcb_get_geometry_unchecked(globalconf.connection, icon);
+    if (mask)
+        geom_mask_c = xcb_get_geometry_unchecked(globalconf.connection, mask);
+    geom_icon_r = xcb_get_geometry_reply(globalconf.connection, geom_icon_c, NULL);
+    if (mask)
+        geom_mask_r = xcb_get_geometry_reply(globalconf.connection, geom_mask_c, NULL);
+
+    if (!geom_icon_r || (mask && !geom_mask_r))
+        goto out;
+    if ((geom_icon_r->depth != 1 && geom_icon_r->depth != globalconf.screen->root_depth)
+            || (geom_mask_r && geom_mask_r->depth != 1))
+    {
+        warn("Got pixmaps with depth (%d, %d) while processing icon, but only depth 1 and %d are allowed",
+                geom_icon_r->depth, geom_mask_r ? geom_mask_r->depth : 0, globalconf.screen->root_depth);
+        goto out;
+    }
+
+    if (geom_icon_r->depth == 1)
+        s_icon = cairo_xcb_surface_create_for_bitmap(globalconf.connection,
+                globalconf.screen, icon, geom_icon_r->width, geom_icon_r->height);
+    else
+        s_icon = cairo_xcb_surface_create(globalconf.connection, icon, globalconf.default_visual,
+                geom_icon_r->width, geom_icon_r->height);
+    result = s_icon;
+
+    if (mask)
+    {
+        cairo_surface_t *s_mask;
+        cairo_t *cr;
+
+        result = cairo_surface_create_similar(s_icon, CAIRO_CONTENT_COLOR_ALPHA, geom_icon_r->width, geom_icon_r->height);
+        s_mask = cairo_xcb_surface_create_for_bitmap(globalconf.connection,
+                globalconf.screen, mask, geom_icon_r->width, geom_icon_r->height);
+        cr = cairo_create(result);
+
+        cairo_set_source_surface(cr, s_icon, 0, 0);
+        cairo_mask_surface(cr, s_mask, 0, 0);
+        cairo_surface_destroy(s_mask);
+        cairo_destroy(cr);
+    }
+
+    client_set_icon(c, result);
+
+    cairo_surface_destroy(result);
+    if (result != s_icon)
+        cairo_surface_destroy(s_icon);
+
+out:
+    p_delete(&geom_icon_r);
+    p_delete(&geom_mask_r);
+}
+
+
 /** Kill a client.
  * \param L The Lua VM state.
  *
@@ -1517,17 +1646,47 @@ client_get_drawable(client_t *c, int x, int y)
     return client_get_drawable_offset(c, &x, &y);
 }
 
+static void
+client_refresh_titlebar_partial(client_t *c, client_titlebar_t bar, int16_t x, int16_t y, uint16_t width, uint16_t height)
+{
+    if(c->titlebar[bar].drawable == NULL
+            || c->titlebar[bar].drawable->pixmap == XCB_NONE
+            || !c->titlebar[bar].drawable->refreshed)
+        return;
+
+    /* Is the titlebar part of the area that should get redrawn? */
+    area_t area = titlebar_get_area(c, bar);
+    if (AREA_LEFT(area) >= x + width || AREA_RIGHT(area) <= x)
+        return;
+    if (AREA_TOP(area) >= y + height || AREA_BOTTOM(area) <= y)
+        return;
+
+    /* Redraw the affected parts */
+    cairo_surface_flush(c->titlebar[bar].drawable->surface);
+    xcb_copy_area(globalconf.connection, c->titlebar[bar].drawable->pixmap, c->frame_window,
+            globalconf.gc, x - area.x, y - area.y, x, y, width, height);
+}
+
+#define HANDLE_TITLEBAR_REFRESH(name, index)                                                \
+static void                                                                                 \
+client_refresh_titlebar_ ## name(client_t *c)                                               \
+{                                                                                           \
+    area_t area = titlebar_get_area(c, index);                                              \
+    client_refresh_titlebar_partial(c, index, area.x, area.y, area.width, area.height);     \
+}
+HANDLE_TITLEBAR_REFRESH(top, CLIENT_TITLEBAR_TOP)
+HANDLE_TITLEBAR_REFRESH(right, CLIENT_TITLEBAR_RIGHT)
+HANDLE_TITLEBAR_REFRESH(bottom, CLIENT_TITLEBAR_BOTTOM)
+HANDLE_TITLEBAR_REFRESH(left, CLIENT_TITLEBAR_LEFT)
+
+/**
+ * Refresh all titlebars that are in the specified rectangle
+ */
 void
-client_refresh(client_t *c)
+client_refresh_partial(client_t *c, int16_t x, int16_t y, uint16_t width, uint16_t height)
 {
     for (client_titlebar_t bar = CLIENT_TITLEBAR_TOP; bar < CLIENT_TITLEBAR_COUNT; bar++) {
-        if (c->titlebar[bar].drawable == NULL || c->titlebar[bar].drawable->surface == NULL)
-            continue;
-
-        area_t area = titlebar_get_area(c, bar);
-        cairo_surface_flush(c->titlebar[bar].drawable->surface);
-        xcb_copy_area(globalconf.connection, c->titlebar[bar].pixmap, c->frame_window,
-                globalconf.gc, 0, 0, area.x, area.y, area.width, area.height);
+        client_refresh_titlebar_partial(c, bar, x, y, width, height);
     }
 }
 
@@ -1537,7 +1696,22 @@ titlebar_get_drawable(lua_State *L, client_t *c, int cl_idx, client_titlebar_t b
     if (c->titlebar[bar].drawable == NULL)
     {
         cl_idx = luaA_absindex(L, cl_idx);
-        drawable_allocator(L, (drawable_refresh_callback *) client_refresh, c);
+        switch (bar) {
+        case CLIENT_TITLEBAR_TOP:
+            drawable_allocator(L, (drawable_refresh_callback *) client_refresh_titlebar_top, c);
+            break;
+        case CLIENT_TITLEBAR_BOTTOM:
+            drawable_allocator(L, (drawable_refresh_callback *) client_refresh_titlebar_bottom, c);
+            break;
+        case CLIENT_TITLEBAR_RIGHT:
+            drawable_allocator(L, (drawable_refresh_callback *) client_refresh_titlebar_right, c);
+            break;
+        case CLIENT_TITLEBAR_LEFT:
+            drawable_allocator(L, (drawable_refresh_callback *) client_refresh_titlebar_left, c);
+            break;
+        default:
+            fatal("Unknown titlebar kind %d\n", (int) bar);
+        }
         c->titlebar[bar].drawable = luaA_object_ref_item(L, cl_idx, -1);
     }
 
@@ -1557,19 +1731,39 @@ titlebar_resize(client_t *c, client_titlebar_t bar, int size)
      * titlebars should keep its current size!) */
     area_t geometry = c->geometry;
     int change = size - c->titlebar[bar].size;
+    int16_t diff_top = 0, diff_bottom = 0, diff_right = 0, diff_left = 0;
     switch (bar) {
     case CLIENT_TITLEBAR_TOP:
+        geometry.height += change;
+        diff_top = change;
+        break;
     case CLIENT_TITLEBAR_BOTTOM:
         geometry.height += change;
+        diff_bottom = change;
         break;
     case CLIENT_TITLEBAR_RIGHT:
+        geometry.width += change;
+        diff_right = change;
+        break;
     case CLIENT_TITLEBAR_LEFT:
         geometry.width += change;
+        diff_left = change;
         break;
     default:
         fatal("Unknown titlebar kind %d\n", (int) bar);
     }
 
+    if(c->size_hints.flags & XCB_ICCCM_SIZE_HINT_P_WIN_GRAVITY)
+    {
+        int16_t diff_x = 0, diff_y = 0;
+        xwindow_translate_for_gravity(c->size_hints.win_gravity,
+                                      diff_left, diff_top,
+                                      diff_right, diff_bottom,
+                                      &diff_x, &diff_y);
+        geometry.x += diff_x;
+        geometry.y += diff_y;
+    }
+
     c->titlebar[bar].size = size;
     client_resize_do(c, geometry, true, false);
 }
@@ -1597,7 +1791,7 @@ HANDLE_TITLEBAR(right, CLIENT_TITLEBAR_RIGHT)
 HANDLE_TITLEBAR(bottom, CLIENT_TITLEBAR_BOTTOM)
 HANDLE_TITLEBAR(left, CLIENT_TITLEBAR_LEFT)
 
-/** Return client geometry.
+/** Return or set client geometry.
  * \param L The Lua VM state.
  * \return The number of elements pushed on stack.
  * \luastack
@@ -1671,6 +1865,13 @@ luaA_client_set_modal(lua_State *L, client_t *c)
     return 0;
 }
 
+static int
+luaA_client_set_maximized(lua_State *L, client_t *c)
+{
+    client_set_maximized(L, -3, luaA_checkboolean(L, -1));
+    return 0;
+}
+
 static int
 luaA_client_set_maximized_horizontal(lua_State *L, client_t *c)
 {
@@ -1780,42 +1981,41 @@ LUA_OBJECT_EXPORT_PROPERTY(client, client_t, sticky, lua_pushboolean)
 LUA_OBJECT_EXPORT_PROPERTY(client, client_t, size_hints_honor, lua_pushboolean)
 LUA_OBJECT_EXPORT_PROPERTY(client, client_t, maximized_horizontal, lua_pushboolean)
 LUA_OBJECT_EXPORT_PROPERTY(client, client_t, maximized_vertical, lua_pushboolean)
+LUA_OBJECT_EXPORT_PROPERTY(client, client_t, startup_id, lua_pushstring)
+
+static int
+luaA_client_get_maximized(lua_State *L, client_t *c)
+{
+    lua_pushboolean(L, client_get_maximized(c));
+    return 1;
+}
 
 static int
 luaA_client_get_content(lua_State *L, client_t *c)
 {
-    xcb_image_t *ximage = xcb_image_get(globalconf.connection,
-                                        c->window,
-                                        0, 0,
-                                        c->geometry.width,
-                                        c->geometry.height,
-                                        ~0, XCB_IMAGE_FORMAT_Z_PIXMAP);
-    cairo_surface_t *surface = NULL;
+    xcb_get_window_attributes_cookie_t cookie;
+    xcb_get_window_attributes_reply_t *attr;
+    cairo_surface_t *surface;
+    int width  = c->geometry.width;
+    int height = c->geometry.height;
 
-    if(ximage)
-    {
-        if(ximage->bpp >= 24)
-        {
-            uint32_t *data = p_new(uint32_t, ximage->width * ximage->height);
+    /* Just the client size without decorations */
+    width  -= c->titlebar[CLIENT_TITLEBAR_LEFT].size + c->titlebar[CLIENT_TITLEBAR_RIGHT].size;
+    height -= c->titlebar[CLIENT_TITLEBAR_TOP].size + c->titlebar[CLIENT_TITLEBAR_BOTTOM].size;
 
-            for(int y = 0; y < ximage->height; y++)
-                for(int x = 0; x < ximage->width; x++)
-                {
-                    data[y * ximage->width + x] = xcb_image_get_pixel(ximage, x, y);
-                    data[y * ximage->width + x] |= 0xff000000; /* set alpha to 0xff */
-                }
+    cookie = xcb_get_window_attributes(globalconf.connection, c->window);
+    attr = xcb_get_window_attributes_reply(globalconf.connection, cookie, NULL);
 
-            surface = draw_surface_from_data(ximage->width, ximage->height, data);
-            p_delete(&data);
-        }
-        xcb_image_destroy(ximage);
-    }
-
-    if (!surface)
+    if (!attr)
         return 0;
 
+    surface = cairo_xcb_surface_create(globalconf.connection, c->window,
+                                       draw_find_visual(globalconf.screen, attr->visual),
+                                       width, height);
+
     /* lua has to make sure to free the ref or we have a leak */
     lua_pushlightuserdata(L, surface);
+    free(attr);
     return 1;
 }
 
@@ -1977,7 +2177,39 @@ luaA_client_get_size_hints(lua_State *L, client_t *c)
     return 1;
 }
 
-/** Set the client's bounding shape.
+/** Get the client's child window bounding shape.
+ * \param L The Lua VM state.
+ * \param client The client object.
+ * \return The number of elements pushed on stack.
+ */
+static int
+luaA_client_get_client_shape_bounding(lua_State *L, client_t *c)
+{
+    cairo_surface_t *surf = xwindow_get_shape(c->window, XCB_SHAPE_SK_BOUNDING);
+    if (!surf)
+        return 0;
+    /* lua has to make sure to free the ref or we have a leak */
+    lua_pushlightuserdata(L, surf);
+    return 1;
+}
+
+/** Get the client's frame window bounding shape.
+ * \param L The Lua VM state.
+ * \param client The client object.
+ * \return The number of elements pushed on stack.
+ */
+static int
+luaA_client_get_shape_bounding(lua_State *L, client_t *c)
+{
+    cairo_surface_t *surf = xwindow_get_shape(c->frame_window, XCB_SHAPE_SK_BOUNDING);
+    if (!surf)
+        return 0;
+    /* lua has to make sure to free the ref or we have a leak */
+    lua_pushlightuserdata(L, surf);
+    return 1;
+}
+
+/** Set the client's frame window bounding shape.
  * \param L The Lua VM state.
  * \param client The client object.
  * \return The number of elements pushed on stack.
@@ -1992,10 +2224,43 @@ luaA_client_set_shape_bounding(lua_State *L, client_t *c)
             c->geometry.width + (c->border_width * 2),
             c->geometry.height + (c->border_width * 2),
             XCB_SHAPE_SK_BOUNDING, surf, -c->border_width);
+    luaA_object_emit_signal(L, -3, "property::shape_bounding", 0);
     return 0;
 }
 
-/** Set the client's clip shape.
+/** Get the client's child window clip shape.
+ * \param L The Lua VM state.
+ * \param client The client object.
+ * \return The number of elements pushed on stack.
+ */
+static int
+luaA_client_get_client_shape_clip(lua_State *L, client_t *c)
+{
+    cairo_surface_t *surf = xwindow_get_shape(c->window, XCB_SHAPE_SK_CLIP);
+    if (!surf)
+        return 0;
+    /* lua has to make sure to free the ref or we have a leak */
+    lua_pushlightuserdata(L, surf);
+    return 1;
+}
+
+/** Get the client's frame window clip shape.
+ * \param L The Lua VM state.
+ * \param client The client object.
+ * \return The number of elements pushed on stack.
+ */
+static int
+luaA_client_get_shape_clip(lua_State *L, client_t *c)
+{
+    cairo_surface_t *surf = xwindow_get_shape(c->frame_window, XCB_SHAPE_SK_CLIP);
+    if (!surf)
+        return 0;
+    /* lua has to make sure to free the ref or we have a leak */
+    lua_pushlightuserdata(L, surf);
+    return 1;
+}
+
+/** Set the client's frame window clip shape.
  * \param L The Lua VM state.
  * \param client The client object.
  * \return The number of elements pushed on stack.
@@ -2008,6 +2273,7 @@ luaA_client_set_shape_clip(lua_State *L, client_t *c)
         surf = (cairo_surface_t *)lua_touserdata(L, -1);
     xwindow_set_shape(c->frame_window, c->geometry.width, c->geometry.height,
             XCB_SHAPE_SK_CLIP, surf, 0);
+    luaA_object_emit_signal(L, -3, "property::shape_clip", 0);
     return 0;
 }
 
@@ -2029,8 +2295,10 @@ luaA_client_keys(lua_State *L)
     {
         luaA_key_array_set(L, 1, 2, keys);
         luaA_object_emit_signal(L, 1, "property::keys", 0);
-        xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, c->frame_window, XCB_BUTTON_MASK_ANY);
-        xwindow_grabkeys(c->frame_window, keys);
+        xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, c->window, XCB_BUTTON_MASK_ANY);
+        xwindow_grabkeys(c->window, keys);
+        if (c->nofocus_window)
+            xwindow_grabkeys(c->nofocus_window, &c->keys);
     }
 
     return luaA_key_array_get(L, 1, keys);
@@ -2185,6 +2453,10 @@ client_class_setup(lua_State *L)
                             NULL,
                             (lua_class_propfunc_t) luaA_client_get_group_window,
                             NULL);
+    luaA_class_add_property(&client_class, "maximized",
+                            (lua_class_propfunc_t) luaA_client_set_maximized,
+                            (lua_class_propfunc_t) luaA_client_get_maximized,
+                            (lua_class_propfunc_t) luaA_client_set_maximized);
     luaA_class_add_property(&client_class, "maximized_horizontal",
                             (lua_class_propfunc_t) luaA_client_set_maximized_horizontal,
                             (lua_class_propfunc_t) luaA_client_get_maximized_horizontal,
@@ -2231,12 +2503,24 @@ client_class_setup(lua_State *L)
                             NULL);
     luaA_class_add_property(&client_class, "shape_bounding",
                             (lua_class_propfunc_t) luaA_client_set_shape_bounding,
-                            NULL,
+                            (lua_class_propfunc_t) luaA_client_get_shape_bounding,
                             (lua_class_propfunc_t) luaA_client_set_shape_bounding);
     luaA_class_add_property(&client_class, "shape_clip",
                             (lua_class_propfunc_t) luaA_client_set_shape_clip,
-                            NULL,
+                            (lua_class_propfunc_t) luaA_client_get_shape_clip,
                             (lua_class_propfunc_t) luaA_client_set_shape_clip);
+    luaA_class_add_property(&client_class, "startup_id",
+                            NULL,
+                            (lua_class_propfunc_t) luaA_client_get_startup_id,
+                            NULL);
+    luaA_class_add_property(&client_class, "client_shape_bounding",
+                            NULL,
+                            (lua_class_propfunc_t) luaA_client_get_client_shape_bounding,
+                            NULL);
+    luaA_class_add_property(&client_class, "client_shape_clip",
+                            NULL,
+                            (lua_class_propfunc_t) luaA_client_get_client_shape_clip,
+                            NULL);
 
     signal_add(&client_class.signals, "focus");
     signal_add(&client_class.signals, "list");
@@ -2259,6 +2543,7 @@ client_class_setup(lua_State *L)
     signal_add(&client_class.signals, "property::instance");
     signal_add(&client_class.signals, "property::keys");
     signal_add(&client_class.signals, "property::machine");
+    signal_add(&client_class.signals, "property::maximized");
     signal_add(&client_class.signals, "property::maximized_horizontal");
     signal_add(&client_class.signals, "property::maximized_vertical");
     signal_add(&client_class.signals, "property::minimized");
@@ -2268,6 +2553,10 @@ client_class_setup(lua_State *L)
     signal_add(&client_class.signals, "property::pid");
     signal_add(&client_class.signals, "property::role");
     signal_add(&client_class.signals, "property::screen");
+    signal_add(&client_class.signals, "property::shape_bounding");
+    signal_add(&client_class.signals, "property::shape_client_bounding");
+    signal_add(&client_class.signals, "property::shape_client_clip");
+    signal_add(&client_class.signals, "property::shape_clip");
     signal_add(&client_class.signals, "property::size_hints_honor");
     signal_add(&client_class.signals, "property::skip_taskbar");
     signal_add(&client_class.signals, "property::sticky");
@@ -2279,9 +2568,11 @@ client_class_setup(lua_State *L)
     signal_add(&client_class.signals, "property::window");
     signal_add(&client_class.signals, "property::x");
     signal_add(&client_class.signals, "property::y");
+    signal_add(&client_class.signals, "request::activate");
     signal_add(&client_class.signals, "request::fullscreen");
     signal_add(&client_class.signals, "request::maximized_horizontal");
     signal_add(&client_class.signals, "request::maximized_vertical");
+    signal_add(&client_class.signals, "request::tag");
     signal_add(&client_class.signals, "tagged");
     signal_add(&client_class.signals, "unfocus");
     signal_add(&client_class.signals, "unmanage");
