@@ -71,8 +71,23 @@ screen_add(screen_t new_screen)
                     MAX(new_screen.geometry.height, screen_to_test->geometry.height);
                 return;
             }
+
     signal_add(&new_screen.signals, "property::workarea");
     screen_array_append(&globalconf.screens, new_screen);
+
+    /* Allocate the lua userdata object representing this screen */
+    screen_t *s = &globalconf.screens.tab[globalconf.screens.len-1];
+    screen_t **ps = lua_newuserdata(globalconf.L, sizeof(*ps));
+    *ps = s;
+    luaL_getmetatable(globalconf.L, "screen");
+    lua_setmetatable(globalconf.L, -2);
+    s->userdata = luaA_object_ref(globalconf.L, -1);
+}
+
+static bool
+screens_exist(void)
+{
+    return globalconf.screens.len > 0;
 }
 
 static bool
@@ -150,7 +165,7 @@ screen_scan_randr(void)
 
             p_delete(&screen_res_r);
 
-            return true;
+            return screens_exist();
         }
     }
 
@@ -195,7 +210,7 @@ screen_scan_xinerama(void)
 
         p_delete(&xsq);
 
-        return true;
+        return screens_exist();
     }
 
     return false;
@@ -392,10 +407,7 @@ screen_client_moveto(client_t *c, screen_t *new_screen, bool doresize)
 static int
 luaA_pushscreen(lua_State *L, screen_t *s)
 {
-    screen_t **ps = lua_newuserdata(L, sizeof(*ps));
-    *ps = s;
-    luaL_getmetatable(L, "screen");
-    lua_setmetatable(L, -2);
+    luaA_object_push(L, s->userdata);
     return 1;
 }
 
@@ -410,7 +422,7 @@ luaA_screen_module_index(lua_State *L)
 {
     const char *name;
 
-    if((name = lua_tostring(L, 2)))
+    if(lua_type(L, 2) == LUA_TSTRING && (name = lua_tostring(L, 2)))
         foreach(screen, globalconf.screens)
             foreach(output, screen->outputs)
                 if(A_STREQ(output->name, name))
@@ -490,24 +502,7 @@ luaA_screen_index(lua_State *L)
         return 1;
     }
 
-    return 0;
-}
-
-/** A screen.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
-static int
-luaA_screen_equal(lua_State *L)
-{
-    screen_t **ps1;
-    screen_t **ps2;
-
-    ps1 = luaL_checkudata(L, 1, "screen");
-    ps2 = luaL_checkudata(L, 2, "screen");
-    lua_pushboolean(L, *ps1 == *ps2);
-
-    return 1;
+    return luaA_default_index(L);
 }
 
 /** Add a signal to a screen.
@@ -615,6 +610,7 @@ const struct luaL_Reg awesome_screen_methods[] =
 {
     { "count", luaA_screen_count },
     { "__index", luaA_screen_module_index },
+    { "__newindex", luaA_default_newindex },
     { NULL, NULL }
 };
 
@@ -625,7 +621,7 @@ const struct luaL_Reg awesome_screen_meta[] =
     { "disconnect_signal", luaA_screen_disconnect_signal },
     { "emit_signal", luaA_screen_emit_signal },
     { "__index", luaA_screen_index },
-    { "__eq", luaA_screen_equal },
+    { "__newindex", luaA_default_newindex },
     { NULL, NULL }
 };
 
