@@ -39,15 +39,36 @@
 void
 systray_init(void)
 {
+    xcb_intern_atom_cookie_t atom_systray_q;
+    xcb_intern_atom_reply_t *atom_systray_r;
+    char *atom_name;
     xcb_screen_t *xscreen = globalconf.screen;
 
     globalconf.systray.window = xcb_generate_id(globalconf.connection);
+    globalconf.systray.background_pixel = xscreen->black_pixel;
     xcb_create_window(globalconf.connection, xscreen->root_depth,
                       globalconf.systray.window,
                       xscreen->root,
                       -1, -1, 1, 1, 0,
                       XCB_COPY_FROM_PARENT, xscreen->root_visual,
-                      0, NULL);
+                      XCB_CW_BACK_PIXEL, (const uint32_t [])
+                      { xscreen->black_pixel });
+
+    atom_name = xcb_atom_name_by_screen("_NET_SYSTEM_TRAY", globalconf.default_screen);
+    if(!atom_name)
+        fatal("error getting systray atom name");
+
+    atom_systray_q = xcb_intern_atom_unchecked(globalconf.connection, false,
+                                               a_strlen(atom_name), atom_name);
+
+    p_delete(&atom_name);
+
+    atom_systray_r = xcb_intern_atom_reply(globalconf.connection, atom_systray_q, NULL);
+    if(!atom_systray_r)
+        fatal("error getting systray atom");
+
+    globalconf.systray.atom = atom_systray_r->atom;
+    p_delete(&atom_systray_r);
 }
 
 /** Register systray in X.
@@ -57,22 +78,11 @@ systray_register(void)
 {
     xcb_client_message_event_t ev;
     xcb_screen_t *xscreen = globalconf.screen;
-    char *atom_name;
-    xcb_intern_atom_cookie_t atom_systray_q;
-    xcb_intern_atom_reply_t *atom_systray_r;
-    xcb_atom_t atom_systray;
 
-    /* Send requests */
-    if(!(atom_name = xcb_atom_name_by_screen("_NET_SYSTEM_TRAY", globalconf.default_screen)))
-    {
-        warn("error getting systray atom");
+    if(globalconf.systray.registered)
         return;
-    }
 
-    atom_systray_q = xcb_intern_atom_unchecked(globalconf.connection, false,
-                                               a_strlen(atom_name), atom_name);
-
-    p_delete(&atom_name);
+    globalconf.systray.registered = true;
 
     /* Fill event */
     p_clear(&ev, 1);
@@ -81,22 +91,13 @@ systray_register(void)
     ev.format = 32;
     ev.type = MANAGER;
     ev.data.data32[0] = XCB_CURRENT_TIME;
+    ev.data.data32[1] = globalconf.systray.atom;
     ev.data.data32[2] = globalconf.systray.window;
     ev.data.data32[3] = ev.data.data32[4] = 0;
 
-    if(!(atom_systray_r = xcb_intern_atom_reply(globalconf.connection, atom_systray_q, NULL)))
-    {
-        warn("error getting systray atom");
-        return;
-    }
-
-    ev.data.data32[1] = atom_systray = atom_systray_r->atom;
-
-    p_delete(&atom_systray_r);
-
     xcb_set_selection_owner(globalconf.connection,
                             globalconf.systray.window,
-                            atom_systray,
+                            globalconf.systray.atom,
                             XCB_CURRENT_TIME);
 
     xcb_send_event(globalconf.connection, false, xscreen->root, 0xFFFFFF, (char *) &ev);
@@ -107,31 +108,16 @@ systray_register(void)
 void
 systray_cleanup(void)
 {
-    xcb_intern_atom_reply_t *atom_systray_r;
-    char *atom_name;
-
-    if(!(atom_name = xcb_atom_name_by_screen("_NET_SYSTEM_TRAY", globalconf.default_screen))
-       || !(atom_systray_r = xcb_intern_atom_reply(globalconf.connection,
-                                                   xcb_intern_atom_unchecked(globalconf.connection,
-                                                                             false,
-                                                                             a_strlen(atom_name),
-                                                                             atom_name),
-                                                   NULL)))
-    {
-        warn("error getting systray atom");
-        p_delete(&atom_name);
+    if(!globalconf.systray.registered)
         return;
-    }
 
-    p_delete(&atom_name);
+    globalconf.systray.registered = false;
 
     xcb_set_selection_owner(globalconf.connection,
                             XCB_NONE,
-                            atom_systray_r->atom,
+                            globalconf.systray.atom,
                             XCB_CURRENT_TIME);
 
-    p_delete(&atom_systray_r);
-
     xcb_unmap_window(globalconf.connection,
                      globalconf.systray.window);
 }
@@ -164,7 +150,6 @@ systray_request_handle(xcb_window_t embed_win, xembed_info_t *info)
 
     xcb_change_window_attributes(globalconf.connection, embed_win, XCB_CW_EVENT_MASK,
                                  select_input_val);
-    xwindow_set_state(embed_win, XCB_ICCCM_WM_STATE_WITHDRAWN);
 
     /* we grab the window, but also make sure it's automatically reparented back
      * to the root window if we should die.
@@ -272,7 +257,7 @@ luaA_systray_invalidate(void)
 }
 
 static void
-systray_update(int base_size, bool horizontal)
+systray_update(int base_size, bool horizontal, bool reverse, bool force_redraw)
 {
     if(base_size <= 0)
         return;
@@ -293,11 +278,19 @@ systray_update(int base_size, bool horizontal)
     config_vals[2] = config_vals[3] = base_size;
     for(int i = 0; i < globalconf.embedded.len; i++)
     {
-        xembed_window_t *em = &globalconf.embedded.tab[i];
+        xembed_window_t *em;
+
+        if(reverse)
+            em = &globalconf.embedded.tab[(globalconf.embedded.len - i - 1)];
+        else
+            em = &globalconf.embedded.tab[i];
+
         xcb_configure_window(globalconf.connection, em->win,
                              XCB_CONFIG_WINDOW_X | XCB_CONFIG_WINDOW_Y | XCB_CONFIG_WINDOW_WIDTH | XCB_CONFIG_WINDOW_HEIGHT,
                              config_vals);
         xcb_map_window(globalconf.connection, em->win);
+        if (force_redraw)
+            xcb_clear_area(globalconf.connection, 1, em->win, 0, 0, 0, 0);
         if(horizontal)
             config_vals[0] += base_size;
         else
@@ -314,10 +307,14 @@ systray_update(int base_size, bool horizontal)
  * \lparam y Y position for the systray.
  * \lparam base_size The size (width and height) each systray item gets.
  * \lparam horiz If true, the systray is horizontal, else vertical
+ * \lparam bg Color of the systray background
+ * \lparam revers If true, the systray icon order will be reversed, else default
  */
 int
 luaA_systray(lua_State *L)
 {
+    systray_register();
+
     if(lua_gettop(L) != 0)
     {
         size_t bg_len;
@@ -327,19 +324,22 @@ luaA_systray(lua_State *L)
         int base_size = luaL_checknumber(L, 4);
         bool horiz = lua_toboolean(L, 5);
         const char *bg = luaL_checklstring(L, 6, &bg_len);
+        bool revers = lua_toboolean(L, 7);
         color_t bg_color;
+        bool force_redraw = false;
 
-        if(color_init_reply(color_init_unchecked(&bg_color, bg, bg_len)))
+        if(color_init_reply(color_init_unchecked(&bg_color, bg, bg_len))
+                && globalconf.systray.background_pixel != bg_color.pixel)
         {
             uint32_t config_back[] = { bg_color.pixel };
+            globalconf.systray.background_pixel = bg_color.pixel;
             xcb_change_window_attributes(globalconf.connection,
                                          globalconf.systray.window,
                                          XCB_CW_BACK_PIXEL, config_back);
+            xcb_clear_area(globalconf.connection, 1, globalconf.systray.window, 0, 0, 0, 0);
+            force_redraw = true;
         }
 
-        if(globalconf.systray.parent == NULL)
-            systray_register();
-
         if(globalconf.systray.parent != w)
             xcb_reparent_window(globalconf.connection,
                                 globalconf.systray.window,
@@ -358,7 +358,7 @@ luaA_systray(lua_State *L)
 
         if(globalconf.embedded.len != 0)
         {
-            systray_update(base_size, horiz);
+            systray_update(base_size, horiz, revers, force_redraw);
             xcb_map_window(globalconf.connection,
                            globalconf.systray.window);
         }
